# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os
import pytest
import tvbwidgets.api as api
from tvbwidgets.core.auth import get_current_token, CLB_AUTH
from tvbwidgets.ui.base_widget import TVBWidget
from tvb.simulator.lab import models, integrators


def test_auth():
    token = "Bla-bla-42"
    os.environ[CLB_AUTH] = token
    assert get_current_token() == token


def test_interpret():
    w = api.PhasePlaneWidget(model=models.Generic2dOscillator(),
                             integrator=integrators.HeunDeterministic())
    w.get_widget()
    assert w is not None

    try:
        api.StorageWidget()
        raise RuntimeError("Expected to fail because EBRAINS token does not exist")
    except RuntimeError:
        pass


def test_add_datatype_not_implemented():
    widget = TVBWidget()

    with pytest.raises(NotImplementedError):
        widget.add_datatype(10)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import configparser
from collections import OrderedDict


def parse_ini_file(ini_file):
    # type: (str) -> OrderedDict
    """
    :param ini_file: Valid file path towards an ini file
    :return: Dictionary after parsing
    """
    config = configparser.ConfigParser()
    config.read(filenames=ini_file, encoding='UTF-8')
    sections = config.sections()
    result = OrderedDict()
    for s in sections:
        result[s] = " "
        keys = config[s]
        for key in keys:
            result[key] = config[s][key]
    return result



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os
import inspect
import weakref
import logging
import logging.config


class LoggerBuilder(object):
    """
    Class taking care of uniform Python logger initialization.
    It uses the Python native logging package.
    It's purpose is just to offer a common mechanism for initializing all modules in a package.
    """

    def __init__(self, config_file_name='logging.conf'):
        """
        Prepare Python logger based on a configuration file.
        :param: config_file_name - name of the logging configuration relative to the current package
        """
        current_folder = os.path.dirname(inspect.getfile(self.__class__))
        config_file_path = os.path.join(current_folder, config_file_name)
        logging.config.fileConfig(config_file_path, disable_existing_loggers=False)
        self._loggers = weakref.WeakValueDictionary()

    def build_logger(self, parent_module):
        """
        Build a logger instance and return it
        """
        self._loggers[parent_module] = logger = logging.getLogger(parent_module)
        return logger

    def set_loggers_level(self, level):
        for logger in self._loggers.values():
            logger.setLevel(level)


# We make sure a single instance of logger-builder is created.
GLOBAL_LOGGER_BUILDER = LoggerBuilder()


def get_logger(parent_module=''):
    """
    Function to retrieve a new Python logger instance for current module.

    :param parent_module: module name for which to create logger.
    """
    return GLOBAL_LOGGER_BUILDER.build_logger(parent_module)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

from tvb.basic.neotraits.api import NArray, HasTraits, Range, List, Attr
from tvb.core.neotraits.h5 import H5File, DataSet, Scalar, Json


class PSEData(HasTraits):
    x_title = Attr(str)
    y_title = Attr(str)
    x_value = Attr(list)
    y_value = Attr(list)
    metrics_names = Attr(list)
    results = NArray()


class PSEStorage(H5File):
    def __init__(self, path):
        super(PSEStorage, self).__init__(path)
        self.x_title = Scalar(PSEData.x_title, self)
        self.y_title = Scalar(PSEData.y_title, self)
        self.x_value = Json(PSEData.x_value, self)
        self.y_value = Json(PSEData.y_value, self)
        self.metrics_names = Json(PSEData.metrics_names, self)
        self.results = DataSet(PSEData.results, self)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#
import json
import os

import numpy
import pytest
import shutil
import tempfile

from tvbwidgets.core.simulator.model_exporters import JSONModelExporter, PythonCodeExporter
from tvbwidgets.ui.phase_plane_widget import PhasePlaneWidget
from tvb.simulator.models.oscillator import Generic2dOscillator, SupHopf
from tvbwidgets.tests.constants import OSCILLATOR_2d_DEFAULT_CONFIG, SUP_HOPF_DEFAULT_PARAMS


@pytest.fixture
def temp_storage():
    """
    returns a temp dir and path to a json file in the temp dir
    after each test the temp dir is removed
    """
    temp_storage = tempfile.mkdtemp()
    file_path = os.path.join(temp_storage)
    yield temp_storage, file_path
    shutil.rmtree(temp_storage)


@pytest.fixture(autouse=True)
def mock_exporter_filename(temp_storage, monkeypatch):
    _storage, file = temp_storage
    json_file = os.path.join(file, 'test_json_file.json')
    py_file = os.path.join(file, 'test_py_file.py')
    monkeypatch.setattr(JSONModelExporter, 'file_path', json_file)
    monkeypatch.setattr(PythonCodeExporter, 'file_path', py_file)


def test_josn_export_config_simple():
    """
    tests that widget exports default model configuration to a json
    """
    wid = PhasePlaneWidget(model=Generic2dOscillator(
        **{k: numpy.array(v) for k, v in OSCILLATOR_2d_DEFAULT_CONFIG.items() if k != 'model'}))
    wid.get_widget()
    default_config = OSCILLATOR_2d_DEFAULT_CONFIG
    wid.export_model_configuration()

    with open(JSONModelExporter.file_path, 'r') as exported_json:
        exported_config = json.loads(exported_json.read())
        assert default_config == exported_config['default_config_key1']
        assert len(exported_config.keys()) == 1


def test_json_export_with_user_defined_configuration_name():
    """
    test that widget exports configuration with a user defined name
    """
    wid = PhasePlaneWidget(model=Generic2dOscillator(
        **{k: numpy.array(v) for k, v in OSCILLATOR_2d_DEFAULT_CONFIG.items() if k != 'model'}))
    wid.get_widget()
    config_name = 'test_config name'
    default_config = OSCILLATOR_2d_DEFAULT_CONFIG
    wid.config_name.value = config_name
    wid.export_model_configuration()
    with open(JSONModelExporter.file_path, 'r') as exported_json:
        exported_config = json.loads(exported_json.read())
        assert default_config == exported_config[config_name]
        assert len(exported_config.keys()) == 1


def test_python_export_with_user_defined_configuration_name():
    """
    test that widget exports configuration with a user defined name as comment
    """
    wid = PhasePlaneWidget(model=SupHopf(
        **{k: numpy.array(v) for k, v in SUP_HOPF_DEFAULT_PARAMS.items() if k != 'model'}))
    wid.get_widget()
    config_name = 'test_config name2'
    expected_instance_code = f'# {config_name}\nimport numpy\nfrom tvb.simulator import models\n' \
                             f'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                             f'model_instance\n\n'
    wid.config_name.value = config_name
    wid.export_type.value = 'Python script'
    wid.export_model_configuration()
    with open(PythonCodeExporter.file_path, 'r') as exported_py:
        exported_config = exported_py.read()
        assert exported_config == expected_instance_code


def test_python_multi_export_with_user_defined_configuration_name():
    """
    test that widget exports configuration with a user defined name as comment
    """
    wid = PhasePlaneWidget(model=SupHopf(
        **{k: numpy.array(v) for k, v in SUP_HOPF_DEFAULT_PARAMS.items() if k != 'model'}))
    wid.get_widget()
    config_name = 'test_config name3'
    expected_instance_code = f'# {config_name}\nimport numpy\nfrom tvb.simulator import models\n' \
                             f'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                             f'model_instance\n\n' \
                             f'# {config_name}\n' \
                             'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                             'model_instance\n\n'

    wid.config_name.value = config_name
    wid.export_type.value = 'Python script'
    wid.export_model_configuration()
    wid.export_model_configuration()
    with open(PythonCodeExporter.file_path, 'r') as exported_py:
        exported_config = exported_py.read()
        assert exported_config == expected_instance_code



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os
import ipywidgets
import ebrains_drive
from ebrains_drive.exceptions import DoesNotExist
from tvbwidgets.core.auth import get_current_token
from tvbwidgets.ui.base_widget import TVBWidget


class DriveWidget(ipywidgets.VBox, TVBWidget):
    ROOT = '/'
    PARENT_DIR = '..'
    DIR_ICON = '\U0001F4C1'

    def __init__(self, collab=None, folder=None, **kwargs):
        TVBWidget.__init__(self, **kwargs)
        self.collab = collab
        self.folder = folder
        bearer_token = get_current_token()
        self.client = ebrains_drive.connect(token=bearer_token)

        try:
            all_repos = self.client.repos.list_repos()
        except Exception:
            self.logger.error("Could not retrieve Repos from EBRAINS Drive!")
            all_repos = []
        dropdown_options = {repo.name: repo for repo in all_repos}

        layout = ipywidgets.Layout(width='400px')
        self.repos_dropdown = ipywidgets.Dropdown(description='Repository', value=dropdown_options.get(self.collab),
                                                  options=dropdown_options, layout=layout)
        self.files_list = ipywidgets.Select(description='Files', value=None, disabled=False, layout=layout)

        self.repos_dropdown.observe(self.select_repo, names='value')
        self.files_list.observe(self.select_dir, names='value')

        self._parent_dir = None
        self._map_names_to_files = dict()

        if self.collab:
            self.select_repo(None, folder=self.folder)
            if self.collab not in list(dropdown_options):
                self.logger.warning(f"Could not find the Collab: {self.collab}. You can manually browse for it.")

        ipywidgets.VBox.__init__(self, [self.repos_dropdown, self.files_list], **kwargs)

    def get_chosen_repo(self):
        return self.repos_dropdown.value

    def get_selected_file_path(self):
        file = self._map_names_to_files.get(self.files_list.value)
        if file is None:
            return file
        return file.path

    def get_selected_file_content(self):
        repo_obj = self.get_chosen_repo()
        file_obj = repo_obj.get_file(self.get_selected_file_path())
        return file_obj.get_content()

    def select_repo(self, _, folder=None):
        if folder is None:
            folder = self.ROOT
        if not folder.startswith(self.ROOT):
            folder = self.ROOT + folder
        self.update_files_for_chosen_dir(folder)

    def select_dir(self, _):
        current_selection = self._map_names_to_files.get(self.files_list.value)
        if not current_selection or not current_selection.isdir:
            return
        self.update_files_for_chosen_dir(current_selection.path)

    def update_files_for_chosen_dir(self, selected_dir):
        self.files_list.unobserve(self.select_dir, names='value')
        try:
            selected_repo = self.get_chosen_repo()
            self.logger.debug("Update Files called with {} and {}".format(selected_repo, selected_dir))
            self._gather_files(selected_repo, selected_dir)
            self.files_list.options = list(self._map_names_to_files.keys())
            self.files_list.value = None
        except Exception as e:
            raise e
        finally:
            self.files_list.observe(self.select_dir, names='value')

    def _gather_files(self, repo, sub_folder):
        dir_icon = self.DIR_ICON

        try:
            dir_obj = repo.get_dir(sub_folder)
            if sub_folder == self.ROOT:
                self._parent_dir = None
            else:
                self._parent_dir = repo.get_dir(os.path.dirname(dir_obj.path))
            self._map_names_to_files = {self.PARENT_DIR: self._parent_dir}

            files_in_repository = dir_obj.ls(force_refresh=True)

            for file in files_in_repository:
                filename = file.name
                if file.isdir:
                    filename = dir_icon + filename
                self._map_names_to_files.update({filename: file})
        except DoesNotExist:
            self.logger.error("Folder {} could not be accessed".format(sub_folder))


class DriveUploadWidget(ipywidgets.HBox, TVBWidget):

    def __init__(self, **kwargs):
        self.repo_browser = DriveWidget()

        self.upload = ipywidgets.FileUpload(description="Choose file", multiple=False)

        self.upload_button = ipywidgets.Button(description="Upload")

        def on_upload_change(s):
            self.logger.debug("Uploading file to drive...")
            upload_value_dict = self.upload.value
            content = list(upload_value_dict.values())[0]['content']
            filename = list(upload_value_dict)[0]
            local_filename = os.path.join(".temp", filename)
            os.makedirs(".temp", exist_ok=True)
            self.logger.debug("Storing local {} ".format(local_filename))
            with open(local_filename, "wb") as fp:
                fp.write(content)

            selected_repo = self.repo_browser.get_chosen_repo()
            selected_dir = self.repo_browser.get_chosen_dir()
            selected_dir = selected_repo.get_dir(selected_dir)

            selected_dir.upload_local_file(local_filename)
            os.remove(local_filename)
            self.logger.info("Finished upload {}!".format(filename))
            self.repo_browser.update_files_for_chosen_repo()

        self.upload_button.on_click(on_upload_change)

        super().__init__([ipywidgets.VBox([self.upload, self.upload_button]),
                          self.repo_browser], **kwargs)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import logging
import os
import numpy
import pytest

import tvbwidgets.api as api
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.region_mapping import RegionMapping
from tvb.datatypes.sensors import SensorsInternal, Sensors
from tvb.datatypes.surfaces import FaceSurface, CorticalSurface, Surface

from tvbwidgets.core.auth import CLB_AUTH
from tvbwidgets.core.exceptions import InvalidFileException
from tvbwidgets.tests.test_drive_widget import MockDriveClient
from tvbwidgets.ui.head_widget import HeadBrowser

NOT_SUPPORTED = 'not supported'


def test_add_datatype(caplog, mocker):
    def mock_update_plot(self):
        """Mock plot drawing"""
        pass

    mocker.patch('tvbwidgets.ui.head_widget.CustomOutput.update_plot', mock_update_plot)

    logger = logging.getLogger('tvbwidgets')
    logger.propagate = True

    connectivity = Connectivity(centres=numpy.zeros((10, 3)))
    widget = api.HeadWidget([connectivity])
    assert widget.output_plot.total_actors == 1
    assert len(widget.plot_controls.children) == 1

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        api.HeadWidget(None)
        assert len(caplog.records) == 0

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        api.HeadWidget('abc')
        assert caplog.records[0].levelname == 'WARNING'
        assert NOT_SUPPORTED in caplog.text

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        api.HeadWidget([10])
        assert caplog.records[0].levelname == 'WARNING'
        assert NOT_SUPPORTED in caplog.text

    widget = api.HeadWidget()

    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        widget.add_datatype(None)
        assert caplog.records[0].levelname == 'INFO'
        assert 'None' in caplog.text

    caplog.clear()
    widget.add_datatype(RegionMapping())
    assert caplog.records[0].levelname == 'INFO'
    assert 'cmap' in caplog.text

    caplog.clear()
    widget.add_datatype(10)
    assert caplog.records[0].levelname == 'WARNING'
    assert NOT_SUPPORTED in caplog.text

    widget.add_datatype(connectivity)
    assert widget.output_plot.total_actors == 1
    assert len(widget.plot_controls.children) == 1

    face = FaceSurface(vertices=numpy.zeros((10, 3)), triangles=numpy.zeros((10, 3), dtype=int))
    widget.add_datatype(face)
    assert widget.output_plot.total_actors == 2
    assert len(widget.plot_controls.children) == 2

    seeg = SensorsInternal(locations=numpy.zeros((10, 3)))
    widget.add_datatype(seeg)
    assert widget.output_plot.total_actors == 3
    assert len(widget.plot_controls.children) == 3

    cortex = CorticalSurface(vertices=numpy.zeros((10, 3)), triangles=numpy.zeros((10, 3), dtype=int))
    reg_map = RegionMapping(array_data=numpy.zeros(10, dtype=int))
    config = api.HeadWidgetConfig(name='Cortex')
    config.add_region_mapping_as_cmap(reg_map)
    widget.add_datatype(cortex)
    assert widget.output_plot.total_actors == 4
    assert len(widget.plot_controls.children) == 4

    left_spots = widget.output_plot.MAX_ACTORS - widget.output_plot.total_actors

    caplog.clear()
    for _ in range(left_spots + 1):
        widget.add_datatype(connectivity)
    assert caplog.records[0].levelname == 'INFO'
    assert 'reached the maximum' in caplog.text

    logger.propagate = False


def test_head_widget(mocker):
    def mockk(token):
        return MockDriveClient()

    mocker.patch('ebrains_drive.connect', mockk)

    if os.environ.get(CLB_AUTH):
        os.environ.pop(CLB_AUTH)

    with pytest.raises(RuntimeError):
        HeadBrowser()

    os.environ[CLB_AUTH] = "test_auth_token"
    widget = HeadBrowser()

    assert len(widget.buttons.children) == 3

    with pytest.raises(InvalidFileException):
        widget._TVBWidgetWithBrowser__validate_file(None, None)

    with pytest.raises(InvalidFileException):
        widget._TVBWidgetWithBrowser__validate_file('abc.txt', '.zip')

    widget._TVBWidgetWithBrowser__display_message('ABC')
    assert widget.message_label.value == HeadBrowser.MSG_TEMPLATE.format('ABC', HeadBrowser.MSG_COLOR)

    widget.storage_widget.api.repos_dropdown.value = widget.storage_widget.api.repos_dropdown.options['repo1']
    widget.storage_widget.api.files_list.value = widget.storage_widget.api.files_list.options[1]

    widget.load_selected_file(Surface)
    assert 'Only .zip' in widget.message_label.value

    widget.load_selected_file(Sensors, '.txt')
    assert 'Could not load' in widget.message_label.value



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import abc
import json
import numpy
import os
import re
from enum import Enum
from tvb.simulator.models import Model
from tvbwidgets import get_logger
from tvbwidgets.core.exceptions import ModelExporterNotFoundError


class ModelConfigurationExports(Enum):
    JSON = 'JSON'
    PYTHON = 'Python script'


def is_valid_file_name(filename):
    # type: (str) -> bool
    """
    checks if a string is valid python file name
    returns even if the file doesn't end with .py
    """
    patterns = [re.compile(r'\w+.py+$'), re.compile(r'\w+$')]
    return any([re.match(p, filename) for p in patterns])


class ABCModelExporter(abc.ABC):
    @abc.abstractmethod
    def __init__(self, model_instance, keys, file_name=None):
        # type: (Model, list[str], str) -> None
        """
        Abstract constructor to be implemented in subclasses
        """
        if file_name:
            self.file_name = file_name

        self.model_instance = model_instance
        self.keys = keys
        # name of the exported configuration is set
        # from outside of class if we need other name than the default
        self.config_name = self.default_config_name
        self.logger = get_logger(self.__class__.__module__)

    @property
    def file_path(self):
        return f'{self.file_name}.{self.file_extension}'

    @property
    @abc.abstractmethod
    def default_config_name(self):
        # type: () -> str
        """
        Abstract property to be implemented in subclasses representing
        default name of the configuration to be exported
        """
        pass

    @abc.abstractmethod
    def do_export(self):
        """
        Abstract method to be implemented in subclasses
        implements export logic
        """
        pass

    @staticmethod
    def sanitize_property(prop):
        # type: (any) -> any
        """
        makes property safe for json serialization while allowing serialization
        of numpy arrays as lists
        """
        if is_jsonable(prop):
            return prop
        if isinstance(prop, numpy.ndarray):
            return prop.tolist()
        return prop.__class__.__name__


class JSONModelExporter(ABCModelExporter):
    """
    Class used to export a model configuration to json format
    """
    file_name = 'model_config'
    file_extension = 'json'

    def __init__(self, model_instance, keys, file_name=None):
        # type: (Model, list[str], str) -> None
        super(JSONModelExporter, self).__init__(model_instance, keys, file_name)

    @property
    def default_config_name(self):
        return 'default_config_key'

    def do_export(self):
        """
        exports model params and model name as json under the key config_name
        if that is different to the default key or the default key plus length of the existing keys
        in the json file + 1
        """
        config_dict = self._read_existing_config()
        config_name = self.config_name if self.config_name != self.default_config_name else self.default_config_name + str(len(config_dict.keys()) + 1)
        config_dict[config_name] = self._prepare_export_values()
        with open(self.file_path, 'w') as f:
            f.write(json.dumps(config_dict))

    def _read_existing_config(self):
        # type: () -> dict
        """
        reads a json file and returns the dict representing the json data or an empty dict
        if the config file or data in it doesn't exist
        """
        file_name = self.file_path
        config = {}
        try:
            with open(file_name, 'r+') as config_file:
                config = json.loads(config_file.read())
        except (FileNotFoundError, json.JSONDecodeError):
            pass
        return config

    def _prepare_export_values(self):
        # type: () -> dict
        """
        returns a json safe dict containing the model params from the exporter keys
        """
        values = {}
        model_dict = self.model_instance.__dict__

        for key in self.keys:
            try:
                values[key] = self.sanitize_property(model_dict[key])
            except KeyError:
                self.logger.warning(
                    f'Parameter {key} will be skipped, it does not exist on {type(self.model_instance)}')

        values['model'] = self.model_instance.__class__.__name__

        return values


class PythonCodeExporter(ABCModelExporter):
    """
    Exporter class to export a model configuration as python code in
    a python file
    """
    file_name = 'model_instances'
    file_extension = 'json'
    numpy_import = 'import numpy'
    module_name = 'models'
    models_import = f'from tvb.simulator import {module_name}'
    instance_var_name = 'model_instance'

    def __init__(self, model_instance, keys, file_name=None):
        # type: (Model, list[str], str) -> None
        super(PythonCodeExporter, self).__init__(model_instance, keys, file_name)

    @property
    def default_config_name(self):
        return 'default config'

    def do_export(self):
        """
        opens a py file and writes code required to instantiate a model with
        the params set on model at the time of export
        """
        # if the file doesn't exist add modules imports
        add_imports = not os.path.exists(self.file_path)

        code = self.get_instance_code(add_imports=add_imports)

        # open the file to append to existing saved model instances
        with open(self.file_path, 'a') as f:
            f.write(code)

    def get_instance_code(self, add_imports=True):
        # type: (bool) -> str
        """
        generates the code required to instantiate a model and returns it as string.
        Optionally can add the imports for model and numpy.
        """
        class_name = self.model_instance.__class__.__name__
        values = f'# {self.config_name}\n'
        if add_imports:
            values += f'{self.numpy_import}\n{self.models_import}\n'

        model_params = self.get_model_params()
        values += f'{self.instance_var_name} = {self.module_name}.{class_name}({model_params})\n'
        values += f'{self.instance_var_name}\n\n'
        return values

    def get_model_params(self):
        # type: () -> str
        """
        Generates a string representing the parameters required to instantiate the
        model set on exporter in the form 'attr=numpy.array(value), ...'
        """
        model_params = ''
        model_instance_dict = self.model_instance.__dict__

        for key in self.keys:
            try:
                model_params += f'{key}=numpy.array({model_instance_dict[key]}),'
            except KeyError:
                self.logger.warning(
                    f'Parameter {key} will be skipped, it does not exist on {type(self.model_instance)}')
        # strip the last comma on params before returning
        return model_params.rstrip(model_params[-1])


def model_exporter_factory(exporter_type, model, keys, export_filename=None):
    # type: (str, Model, list[str], str) -> ABCModelExporter
    """
    Factory for model exporter creation
    """

    try:
        exporter = ModelConfigurationExports(exporter_type)
    except ValueError:
        raise ModelExporterNotFoundError('Could not find any exporter of the selected type!')

    if exporter == ModelConfigurationExports.PYTHON:
        return PythonCodeExporter(model, keys, export_filename)

    return JSONModelExporter(model, keys, export_filename)


def is_jsonable(x):
    # type: (any) -> bool
    """
    checks if an object is serializable
    """
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import json
import os
import numpy
import pytest
import shutil
import tempfile
from tvbwidgets.core.simulator.model_exporters import is_jsonable, is_valid_file_name, JSONModelExporter, PythonCodeExporter
from tvb.simulator.models.oscillator import Generic2dOscillator, SupHopf
from tvbwidgets.tests.constants import SUP_HOPF_DEFAULT_PARAMS, OSCILLATOR_2d_DEFAULT_CONFIG


def test_is_jsonable():
    """
    tests that function works for serializable objects and fails for non-serializable objects
    """
    text = 'some text'
    number = 123.123
    some_list = [1, 1.23, -123]
    some_dict = {'text': text, 'number': number, 'some_list': some_list}
    assert is_jsonable(text)
    assert is_jsonable(number)
    assert is_jsonable(some_list)
    assert is_jsonable(some_dict)
    assert not is_jsonable(SupHopf)
    assert not is_jsonable(Generic2dOscillator())


def test_is_valid_file_name():
    assert is_valid_file_name('some_py_file')
    assert is_valid_file_name('some_py_file.py')
    assert is_valid_file_name('somefile')
    assert not is_valid_file_name('some_py_file/')
    assert not is_valid_file_name('some_,py_file')


@pytest.fixture
def temp_storage():
    temp_storage = tempfile.mkdtemp()
    file_path = os.path.join(temp_storage)
    yield temp_storage, file_path
    shutil.rmtree(temp_storage)


class TestJSONModelExporter:

    @pytest.fixture(autouse=True)
    def mock_exported_filename(self, temp_storage, monkeypatch):
        _storage, file = temp_storage
        file_json_path = os.path.join(file, 'test_json.json')
        monkeypatch.setattr(JSONModelExporter, 'file_path', file_json_path)

    def _export_sup_hopf_default(self):
        model = SupHopf(a=numpy.array(SUP_HOPF_DEFAULT_PARAMS['a']),
                        omega=numpy.array(SUP_HOPF_DEFAULT_PARAMS['omega']))
        keys = ['a', 'omega']
        exporter = JSONModelExporter(model, keys)
        exporter.do_export()
        return exporter

    def test_simple_sup_hopf_export(self):
        exporter = self._export_sup_hopf_default()
        with open(exporter.file_path, 'r') as file:
            json_content = json.loads(file.read())
            assert json_content[exporter.default_config_name + '1'] == SUP_HOPF_DEFAULT_PARAMS

    def test_multiple_sup_hopf_exports(self):
        exporter = self._export_sup_hopf_default()
        exporter.do_export()
        with open(exporter.file_path, 'r') as file:
            json_content = json.loads(file.read())
            assert len(json_content.keys()) == 2
            assert json_content[exporter.default_config_name + '1'] == SUP_HOPF_DEFAULT_PARAMS
            assert json_content[exporter.default_config_name + '2'] == SUP_HOPF_DEFAULT_PARAMS

    def test_multiple_different_exports(self):
        exporter = self._export_sup_hopf_default()
        model_2 = SupHopf(a=numpy.array([-0.2]),
                          omega=numpy.array(SUP_HOPF_DEFAULT_PARAMS['omega']))
        exporter.model_instance = model_2
        exporter.do_export()
        with open(exporter.file_path, 'r') as file:
            json_content = json.loads(file.read())
            assert len(json_content.keys()) == 2
            assert json_content[exporter.default_config_name + '1'] == SUP_HOPF_DEFAULT_PARAMS
            copy_config = SUP_HOPF_DEFAULT_PARAMS.copy()
            copy_config['a'] = [-0.2]
            assert json_content[exporter.default_config_name + '2'] == copy_config

    def test_multiple_exports_different_models(self):
        exporter = self._export_sup_hopf_default()
        generic_defaults = {key: numpy.array(value) for key, value in OSCILLATOR_2d_DEFAULT_CONFIG.items() if
                            key != 'model'}
        exporter.model_instance = Generic2dOscillator(**generic_defaults)
        exporter.keys = generic_defaults.keys()
        exporter.do_export()
        with open(exporter.file_path, 'r') as file:
            json_content = json.loads(file.read())
            assert json_content[exporter.default_config_name + '1'] == SUP_HOPF_DEFAULT_PARAMS
            assert json_content[exporter.default_config_name + '2'] == OSCILLATOR_2d_DEFAULT_CONFIG


class TestPythonCodeExporter:
    @pytest.fixture(autouse=True)
    def mock_exported_filename(self, temp_storage, monkeypatch):
        _storage, file = temp_storage
        py_file_path = os.path.join(file, 'instances_test_python.py')
        monkeypatch.setattr(PythonCodeExporter, 'file_path', py_file_path)

    def test_simple_sup_hopf_python_export(self):
        args = {k: numpy.array(v) for k, v in SUP_HOPF_DEFAULT_PARAMS.items() if k != 'model'}
        model_instance = SupHopf(**args)
        exporter = PythonCodeExporter(model_instance, args.keys())
        exporter.do_export()
        expected = '# default config\nimport numpy\nfrom tvb.simulator import models\n' \
                   'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                   'model_instance\n\n'
        with open(exporter.file_path, 'r') as file:
            py_content = file.read()
            assert py_content == expected

    def test_multiple_sup_hopf_py_export(self):
        args = {k: numpy.array(v) for k, v in SUP_HOPF_DEFAULT_PARAMS.items() if k != 'model'}
        model_instance = SupHopf(**args)
        exporter = PythonCodeExporter(model_instance, args.keys())
        exporter.do_export()
        exporter.do_export()
        expected = '# default config\nimport numpy\nfrom tvb.simulator import models\n' \
                   'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                   'model_instance\n\n'
        expected += '# default config\n' \
                    'model_instance = models.SupHopf(a=numpy.array([-0.5]),omega=numpy.array([1.]))\n' \
                    'model_instance\n\n'

        with open(exporter.file_path, 'r') as file:
            py_content = file.read()
            assert py_content == expected



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#
"""
A collection of parameter related classes and functions.
- Temporary copy from tvb-inversion package

.. moduleauthor:: Fousek Jan <jan.fousek@univ-amu.fr>
"""

from copy import deepcopy
from typing import List, Any, Optional
from dataclasses import dataclass
import numpy as np
from tvb.analyzers.metric_variance_global import compute_variance_global_metric
from tvb.analyzers.metric_kuramoto_index import compute_kuramoto_index_metric
from tvb.analyzers.metric_proxy_metastability import compute_proxy_metastability_metric
from tvb.analyzers.metric_variance_of_node_variance import compute_variance_of_node_variance_metric
from tvb.datatypes.time_series import TimeSeries
from tvb.simulator.simulator import Simulator
import os
from joblib import Parallel, delayed
import logging
from tvbwidgets.core.pse.pse_data import PSEData, PSEStorage
log = logging.getLogger(__name__)

try:
    from dask.distributed import Client
except ImportError:
    log.info("ImportError: Dask dependency is not included, so this functionality won't be available")
    Client = object


class ParamGetter:
    pass


@dataclass
class SimSeq:
    "A sequence of simulator configurations."
    template: Simulator
    params: List[str]
    values: List[List[Any]]
    getters: Optional[List[Optional[ParamGetter]]] = None  # is the first Optional needed?

    # TODO consider transpose, so a param can have a remote data source
    # to load when constructing the sequence

    def __iter__(self):
        self.pos = 0
        return self

    def __post_init__(self):
        self.template.configure()  # deepcopy doesn't work on un-configured simulator o_O
        if self.getters is None:
            self.getters = [None] * len(self.params)
        else:
            assert len(self.getters) == len(self.params)

    def __next__(self):
        if self.pos >= len(self.values):
            raise StopIteration
        obj = deepcopy(self.template)
        updates = zip(self.params, self.getters, self.values[self.pos])
        for key, getter, val in updates:
            if getter is not None:
                val = getter(val)
            exec(f'obj.{key} = val',
                 {'obj': obj, 'val': val})
        self.pos += 1
        return obj


class Metric:
    "A summary statistic for a simulation."

    def __call__(self, t, y) -> np.ndarray:  # what about multi metric returning dict of statistics? Also, chaining?
        pass


class NodeVariability(Metric):
    "A simplistic simulation statistic."

    def __call__(self, t, y):
        return np.std(y[t > (t[-1] / 2), 0, :, 0], axis=0)


class GlobalVariance(Metric):

    def __init__(self, sample_period, start_point=500, segment=4):
        self.sample_period = sample_period
        self.start_point = start_point
        self.segment = segment

    def __call__(self, t, y):
        ts = TimeSeries(sample_period=self.sample_period)
        ts.data = y
        return compute_variance_global_metric({"time_series": ts, "start_point": self.start_point,
                                               "segment": self.segment})


class KuramotoIndex(Metric):

    def __init__(self, sample_period):
        self.sample_period = sample_period

    def __call__(self, t, y):
        ts = TimeSeries(sample_period=self.sample_period)
        ts.data = y
        return compute_kuramoto_index_metric({"time_series": ts})


class ProxyMetastabilitySynchrony(Metric):

    def __init__(self, mode, sample_period, start_point=500, segment=4):
        self.mode = mode
        self.sample_period = sample_period
        self.start_point = start_point
        self.segment = segment

    def __call__(self, t, y):
        ts = TimeSeries(sample_period=self.sample_period)
        ts.data = y
        return compute_proxy_metastability_metric({"time_series": ts, "start_point": self.start_point,
                                                   "segment": self.segment})[self.mode]


class VarianceNodeVariance(Metric):

    def __init__(self, sample_period, start_point=500, segment=4):
        self.sample_period = sample_period
        self.start_point = start_point
        self.segment = segment

    def __call__(self, t, y):
        ts = TimeSeries(sample_period=self.sample_period)
        ts.data = y
        return compute_variance_of_node_variance_metric({"time_series": ts, "start_point": self.start_point,
                                                         "segment": self.segment})


METRICS = ['GlobalVariance', 'KuramotoIndex', 'ProxyMetastabilitySynchrony Metastability',
           'ProxyMetastabilitySynchrony Synchrony', 'VarianceNodeVariance']


class Reduction:
    pass


@dataclass
class SaveMetricsToDisk(Reduction):
    filename: str

    def __call__(self, metrics_mat: np.ndarray) -> None:
        np.save(self.filename, metrics_mat)


# or save to a bucket or do SNPE then to a bucket, etc.

@dataclass
class SaveDataToDisk(Reduction):
    param1: str
    param2: str
    x_values: list
    y_values: list
    metrics: list
    file_name: str

    def __call__(self, metric_data: np.ndarray) -> None:
        metrics_data_np = np.array(metric_data)
        pse_result = PSEData()
        pse_result.x_title = self.param1
        pse_result.y_title = self.param2

        if self.param1 == "connectivity":
            id_values = [val.title[0:25] + "..." for val in self.x_values]
            pse_result.x_value = id_values
        else:
            pse_result.x_value = self.x_values
        if self.param2 == "connectivity":
            id_values = [val.title[0:25] + "..." for val in self.y_values]
            pse_result.y_value = id_values
        else:
            pse_result.y_value = self.y_values

        pse_result.metrics_names = self.metrics
        pse_result.results = metrics_data_np.reshape(len(self.metrics), len(self.x_values), len(self.y_values))
        if ".h5" not in self.file_name:
            self.file_name += ".h5"

        f = PSEStorage(self.file_name)
        f.store(pse_result)
        log.info(str(self.file_name) + " file created")
        f.close()


@dataclass
class PostProcess:
    metrics: List[Metric]
    reduction: Reduction


class Exec:
    pass


@dataclass
class JobLibExec:
    seq: SimSeq
    post: PostProcess
    backend: Optional[Any]
    checkpoint_dir: Optional[str]

    def _checkpoint(self, result, i):
        if self.checkpoint_dir is not None:
            np.save(os.path.join(self.checkpoint_dir, f'{i}.npy'), result)

    def _load_checkpoint(self, i):
        if self.checkpoint_dir is None:
            return None
        checkpoint_file = os.path.join(self.checkpoint_dir, f'{i}.npy')
        if not os.path.exists(checkpoint_file):
            return None
        result = np.load(checkpoint_file, allow_pickle=True)
        return result

    def _init_checkpoint(self):
        if self.checkpoint_dir is not None:
            if os.path.exists(self.checkpoint_dir):
                log.info(f"Reusing existing checkpoint dir {self.checkpoint_dir}")
                # TODO consistency check
            else:
                os.mkdir(self.checkpoint_dir)
                np.savetxt(os.path.join(self.checkpoint_dir, 'params.txt'), self.seq.params, fmt='%s')
                np.save(os.path.join(self.checkpoint_dir, 'param_vals.npy'), self.seq.values)

    def __call__(self, n_jobs=-1):
        log.info("Simulation starts")
        self._init_checkpoint()
        pool = Parallel(n_jobs)

        @delayed
        def job(sim, i):
            result = self._load_checkpoint(i)
            if result is None:
                if self.backend is not None:
                    runner = self.backend()
                    (t, y), = runner.run_sim(sim.configure())
                else:
                    (t, y), = sim.configure().run()
                result = np.hstack([m(t, y) for m in self.post.metrics])
                self._checkpoint(result, i)
            return result

        metrics = pool(job(_, i) for i, _ in enumerate(self.seq))
        self.post.reduction(metrics)
        log.info("Local launch finished")


@dataclass
class DaskExec(JobLibExec):

    def __call__(self, client: Client):
        self._init_checkpoint()

        checkpoint_dir = self.checkpoint_dir
        if checkpoint_dir is not None:
            checkpoint_dir = os.path.abspath(checkpoint_dir)

        def _checkpoint(result, i):
            if checkpoint_dir is not None:
                np.save(os.path.join(checkpoint_dir, f'{i}.npy'), result)

        def _load_checkpoint(i):
            if checkpoint_dir is None:
                return None
            checkpoint_file = os.path.join(checkpoint_dir, f'{i}.npy')
            if not os.path.exists(checkpoint_file):
                return None
            result = np.load(checkpoint_file, allow_pickle=True)
            return result

        def job(i, sim):
            result = _load_checkpoint(i)
            if result is None:
                if self.backend is not None:
                    runner = self.backend()
                    (t, y), = runner.run_sim(sim.configure())
                else:
                    (t, y), = sim.configure().run()
                result = np.hstack([m(t, y) for m in self.post.metrics])
                _checkpoint(result, i)
            return result

        def reduction(vals):
            return self.post.reduction(vals)

        metrics = client.map(job, *list(zip(*enumerate(self.seq))))

        if self.post.reduction is not None:
            reduced = client.submit(reduction, metrics)
            return reduced.result()
        else:
            return metrics


def compute_metrics(sim, metrics):
    computed_metrics = []

    for metric in metrics:
        if metric == "GlobalVariance":
            resulted_metric = (GlobalVariance(sim.monitors[0].period))
        elif metric == "KuramotoIndex":
            resulted_metric = (KuramotoIndex(sim.monitors[0].period))
        elif metric == "ProxyMetastabilitySynchrony Metastability":
            resulted_metric = (ProxyMetastabilitySynchrony("Metastability", sim.monitors[0].period))
        elif metric == "ProxyMetastabilitySynchrony Synchrony":
            resulted_metric = (ProxyMetastabilitySynchrony("Synchrony", sim.monitors[0].period))
        else:
            resulted_metric = (VarianceNodeVariance(sim.monitors[0].period))
        computed_metrics.append(resulted_metric)

    return computed_metrics


def launch_local_param(simulator, param1, param2, x_values, y_values, metrics, file_name):
    input_values = []
    for elem1 in x_values:
        for elem2 in y_values:
            if param1 == "conduction_speed" or param1 == "connectivity":
                el1_value = elem1
            else:
                el1_value = np.array([elem1])
            if param2 == "conduction_speed" or param2 == "connectivity":
                el2_value = elem2
            else:
                el2_value = np.array([elem2])
            input_values.append([el1_value, el2_value])

    sim = simulator.configure()  # deepcopy doesn't work on un-configured simulator o_O
    seq = SimSeq(
        template=sim,
        params=[param1, param2],
        values=input_values
    )
    pp = PostProcess(
        metrics=compute_metrics(sim, metrics),
        reduction=SaveDataToDisk(param1, param2, x_values, y_values, metrics, file_name),
    )
    exe = JobLibExec(seq=seq, post=pp, backend=None, checkpoint_dir=None)
    exe(n_jobs=4)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os.path
import plotly.graph_objects as go
import ipywidgets as widgets
from IPython.core.display_functions import display
from tvbwidgets.core.exceptions import InvalidInputException
from tvbwidgets.core.pse.pse_data import PSEData, PSEStorage
from tvbwidgets.ui.base_widget import TVBWidget


class PSEWidget(TVBWidget):
    """Visualize PSE results"""

    def __init__(self, file_name, **kwargs):
        # type: (str, dict) -> None
        """
        :param file_name: path to the file_name that contains the data necessary for the visualization
        """
        super().__init__(**kwargs)
        if not os.path.exists(file_name):
            raise InvalidInputException(f"File {file_name} was not found!")
        self.file_name = file_name
        self.param1_title = None
        self.param2_title = None
        self.param1_value = None
        self.param2_value = None
        self.data = None
        self.metrics_names = []
        self.dict_metrics = {}
        self.figure = None
        self.smooth_effect_cb = None
        self.change_color_dd = None
        self.metrics_change_dd = None
        self.read_h5_file()
        self._map_names_to_metrics()
        self._create_visualizer()

    def read_h5_file(self):
        pse_result = PSEData()
        PSEStorage(self.file_name).load_into(pse_result)
        self.param1_title = pse_result.x_title
        self.param2_title = pse_result.y_title
        self.param1_value = pse_result.x_value
        self.param2_value = pse_result.y_value
        self.metrics_names = pse_result.metrics_names
        self.data = pse_result.results

    def _map_names_to_metrics(self):
        for index in range(self.metrics_names.__len__()):
            self.dict_metrics[self.metrics_names[index]] = self.data[index]

    def _create_visualizer(self):
        self.param1_value = [str(elem) for elem in self.param1_value]
        self.param2_value = [str(elem) for elem in self.param2_value]
        pse_layout = go.Layout(width=1000, height=500,
                               xaxis=go.layout.XAxis(linecolor='black', linewidth=1, mirror=True,
                                                     title=self.param2_title),
                               yaxis=go.layout.YAxis(linecolor='black', linewidth=1, mirror=True,
                                                     title=self.param1_title),
                               margin=go.layout.Margin(l=100, r=50, b=100, t=100, pad=4), title="PSE Visualizer",
                               titlefont=dict(size=20, family='Arial, sans-serif'), )
        self.figure = go.FigureWidget(layout=pse_layout)

        self.figure.add_trace(go.Heatmap(z=list(self.dict_metrics.values())[0], x=self.param2_value, y=self.param1_value
                                         , colorscale='RdBu', connectgaps=False, showscale=True, zsmooth='best'))

        self._populate_features()

    def _populate_features(self):
        self._smooth_effect()
        self._colors_options()
        self._metrics_options()
        features_vbox = widgets.VBox(children=[self.smooth_effect_cb, self.change_color_dd,
                                               self.metrics_change_dd])
        features_accordion = widgets.Accordion(children=[features_vbox], selected_index=None,
                                               layout=widgets.Layout(width='25%', marginTop='100px'))
        features_accordion.set_title(0, 'Features')
        table = widgets.HBox([features_accordion, self.figure], layout=self.DEFAULT_BORDER)
        display(table)

    def _smooth_effect(self):
        self.smooth_effect_cb = widgets.Checkbox(value=True, description='Smooth visualizer',
                                                 layout=widgets.Layout(margin='10px 0px 10px 0px'))

        def smooth_effect_changed(change):
            if change['type'] != 'change' or change['name'] != 'value':
                return

            if change['new']:
                effect = 'best'
            else:
                effect = False
            self.figure.update_traces(go.Heatmap(zsmooth=effect))

        self.smooth_effect_cb.observe(smooth_effect_changed)

    def _colors_options(self):
        self.change_color_dd = widgets.Dropdown(
            options=['Blackbody', 'Earth', 'Jet', 'Picnic', 'RdBu', 'Rainbow'],
            description="Color:",
            value='RdBu',
            disabled=False)

        def color_changed(change):
            if change['type'] != 'change' or change['name'] != 'value':
                return

            self.figure.update_traces(go.Heatmap(colorscale=change['new']))

        self.change_color_dd.observe(color_changed)

    def _metrics_options(self):
        self.metrics_change_dd = widgets.Dropdown(
            options=self.metrics_names,
            description="Metric:",
            value=list(self.dict_metrics.keys())[0],
            disabled=False)

        def metric_changed(change):
            if change['type'] != 'change' or change['name'] != 'value':
                return

            self.figure.update_traces(go.Heatmap(name=change['new'], z=self.dict_metrics[change['new']]))

        self.metrics_change_dd.observe(metric_changed)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import json
from pathlib import Path
from pkg_resources import get_distribution, DistributionNotFound
from .core.logger.builder import get_logger

LOGGER = get_logger(__name__)


def _fetch_version():
    here = Path(__file__).parent.parent.resolve()
    print(here)

    for settings in here.rglob("package.json"):
        try:
            with settings.open() as f:
                version = json.load(f)["version"]
                return (
                    version.replace("-alpha.", "a")
                        .replace("-beta.", "b")
                        .replace("-rc.", "rc")
                )
        except FileNotFoundError:
            pass

    raise FileNotFoundError(f"Could not find package.json under dir {here!s}")


try:
    __version__ = get_distribution("tvb-widgets").version
except DistributionNotFound:
    LOGGER.debug("Package is not fully installed")
    try:
        __version__ = _fetch_version()

        LOGGER.debug("Version read from the internal package.json file")
    except FileNotFoundError:
        LOGGER.warn("Version not found, we will use fallback")
        __version__ = "1.0"

LOGGER.info(f"Version: {__version__}")



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

from ipywidgets import DOMWidget
from tvb.basic.neotraits.api import HasTraits
from tvbwidgets.core.logger.builder import get_logger


class TVBWidget(object):
    DEFAULT_BORDER = {'border': '2px solid lightgray',
                      'padding': '10px'}

    BUTTON_STYLE = {'height': '40px',
                    'width': '150px'}

    def __init__(self, **kwargs):
        self.logger = get_logger(self.__class__.__module__)

    def get_widget(self):
        if isinstance(self, DOMWidget):
            return self
        self.logger.error("Not a valid widget! Try to overwrite get_widget or inherit DOMWidget!")
        raise RuntimeWarning("Not a valid widget!")

    def add_datatype(self, datatype):
        # type: (HasTraits) -> None
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import ipywidgets
import numpy
import pyvista

from pyvista import PolyData

from tvb.basic.neotraits.api import HasTraits
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.region_mapping import RegionMapping
from tvb.datatypes.sensors import Sensors
from tvb.datatypes.surfaces import Surface

from tvbwidgets.ui.base_widget import TVBWidget
from tvbwidgets.ui.widget_with_browser import TVBWidgetWithBrowser

pyvista.set_jupyter_backend('pythreejs')


class HeadWidgetConfig:

    def __init__(self, name='Actor', style='Surface', color='White', light=True, size=1500, cmap=None, scalars=None):
        self.name = name
        self.style = style
        self.color = color
        self.light = light
        self.size = size
        self.cmap = cmap
        self.scalars = scalars
        self.widget = None

    def add_region_mapping_as_cmap(self, region_mapping):
        # type: (RegionMapping) -> None
        self.scalars = region_mapping.array_data
        self.cmap = 'fire'

    def is_incompatible(self, prev_config):
        # type: (HeadWidgetConfig) -> bool
        if prev_config is None:
            return self.cmap is not None
        if (self.cmap is not None and prev_config.cmap is None) or (self.cmap is None and prev_config.cmap is not None):
            return True
        return False


class CustomOutput(ipywidgets.Output):
    CONFIG = HeadWidgetConfig()
    MAX_ACTORS = 10

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.plotter = pyvista.Plotter()
        self.total_actors = 0

    @property
    def can_draw(self):
        return self.total_actors < self.MAX_ACTORS

    def add_mesh(self, mesh, config=CONFIG):
        if config.cmap is None or config.scalars is None:
            actor = self.plotter.add_mesh(mesh, name=config.name, style=config.style, color=config.color,
                                          lighting=config.light, render=False)
        else:
            actor = self.plotter.add_mesh(mesh, name=config.name, style=config.style, scalars=config.scalars,
                                          cmap=config.cmap, lighting=config.light, render=False)
        self.total_actors += 1
        return actor

    def add_points(self, points, config=CONFIG):
        actor = self.plotter.add_points(points, name=config.name, color=config.color,
                                        point_size=config.size, render=False)
        self.total_actors += 1
        return actor

    def display_actor(self, actor):
        self.plotter.renderer.add_actor(actor, render=False)

    def hide_actor(self, actor):
        self.plotter.renderer.remove_actor(actor, render=False)

    def update_plot(self):
        with self:
            self.clear_output(wait=True)
            self.plotter.show()


class HeadWidget(ipywidgets.HBox, TVBWidget):

    def __init__(self, datatypes=None, ignore=False):
        # type: (list[HasTraits], bool) -> None
        """
        :param datatypes: like Surface, Connectivity, Sensors
        :param ignore: - optional flag which allow ignoring coloring incompatibilities
                       - default is set to False because Ebrains Lab does not support surfaces with incompatible coloring
                        (kills the kernel)
                       - locally this issue does not appear
        """
        self.output_plot = CustomOutput()
        self.plot_controls = ipywidgets.Accordion(layout=ipywidgets.Layout(width='380px'))
        self.existent_configs = []
        self.ignore = ignore

        super().__init__([self.plot_controls, self.output_plot], layout=self.DEFAULT_BORDER)

        if datatypes is not None:
            if not isinstance(datatypes, list):
                self.logger.warning("Input not supported. Please provide a list of datatypes.")
            else:
                for datatype in datatypes:
                    self.add_datatype(datatype)

    def add_datatype(self, datatype, config=None):
        # type: (HasTraits, HeadWidgetConfig) -> None
        if datatype is None:
            self.logger.info("The provided datatype is None!")
            return

        if self.output_plot.can_draw is False:
            self.logger.info("You have reached the maximum datatypes that can be drawn to this plot!")
            return

        if isinstance(datatype, Surface):
            self.__draw_mesh_actor(datatype, config)
        elif isinstance(datatype, Connectivity):
            self.__draw_connectivity_actor(datatype, config)
        elif isinstance(datatype, Sensors):
            self.__draw_sensors_actor(datatype, config)
        elif isinstance(datatype, RegionMapping):
            self.logger.info("RegionMapping should be given as cmap in the config parameter!")
        else:
            self.logger.warning(f"Datatype {type(datatype)} not supported by this widget!")

    def __prepare_mesh(self, surface):
        # type: (Surface) -> PolyData
        dim_4th = numpy.full((surface.triangles.shape[0], 1), 3, dtype=int)
        faces = numpy.hstack((dim_4th, surface.triangles))

        mesh = PolyData(surface.vertices, faces)
        return mesh

    def __toggle_actor(self, change, actor):
        if change.type == 'change':
            if change.new is True:
                self.output_plot.display_actor(actor)
            else:
                self.output_plot.hide_actor(actor)
            self.output_plot.update_plot()

    def __draw_mesh_actor(self, surface, config):
        # type: (Surface, HeadWidgetConfig) -> None

        if config is None:
            config = HeadWidgetConfig(name='Surface-' + str(surface.number_of_vertices))

        if not self.ignore:
            for prev_config in self.existent_configs:
                if config.is_incompatible(prev_config):
                    self.logger.error("HeadWidget can not support multiple surfaces with incompatible coloring!")
                    return
        self.existent_configs.append(config)

        mesh = self.__prepare_mesh(surface)
        mesh_actor = self.output_plot.add_mesh(mesh, config)

        controls_vbox = self._prepare_generic_controls(mesh_actor, config)

        extra_controls = self.__prepare_surface_controls(mesh_actor)
        controls_vbox.children += extra_controls

        self.plot_controls.children += controls_vbox,
        self.output_plot.update_plot()

    def __draw_connectivity_actor(self, connectivity, config):
        # type: (Connectivity, HeadWidgetConfig) -> None
        if config is None:
            config = HeadWidgetConfig(name='Connectivity-' + str(connectivity.number_of_regions), color='Green')

        conn_actor = self.output_plot.add_points(connectivity.centres, config)
        controls_vbox = self._prepare_generic_controls(conn_actor, config)
        extra_controls = self.__prepare_points_controls(conn_actor, config)
        controls_vbox.children += extra_controls

        self.plot_controls.children += controls_vbox,
        self.output_plot.update_plot()

    def __draw_sensors_actor(self, sensors, config):
        # type: (Sensors, HeadWidgetConfig) -> None
        if config is None:
            config = HeadWidgetConfig(name='Sensors-' + str(sensors.number_of_sensors), color='Pink', size=1000)

        sensors_actor = self.output_plot.add_points(sensors.locations, config)
        controls_vbox = self._prepare_generic_controls(sensors_actor, config)
        extra_controls = self.__prepare_points_controls(sensors_actor, config)
        controls_vbox.children += extra_controls

        self.plot_controls.children += controls_vbox,
        self.output_plot.update_plot()

    def _prepare_generic_controls(self, actor, config):
        toggle_prefix = "Toggle "
        title_suffix = " Controls"

        idx = self.output_plot.total_actors - 1
        self.plot_controls.set_title(idx, config.name + title_suffix)

        def toggle_actor(change):
            self.__toggle_actor(change, actor)

        toggle_input = ipywidgets.Checkbox(description=toggle_prefix + config.name, value=True)
        toggle_input.observe(toggle_actor, names=['value'])

        def on_name_change(change):
            value = change['new']
            config.name = value
            toggle_input.description = toggle_prefix + config.name
            self.plot_controls.set_title(idx, config.name + title_suffix)

        name_input = ipywidgets.Text(value=config.name, description='Name: ', disabled=False,
                                     layout=ipywidgets.Layout(width='250px'))
        name_input.observe(on_name_change, names='value')

        def on_color_change(change):
            value = change['new']
            rgb = pyvista.Color(value).float_rgb
            actor.GetProperty().SetColor(rgb[0], rgb[1], rgb[2])
            self.output_plot.update_plot()

        color_input = ipywidgets.ColorPicker(concise=False, description='Color: ', value=config.color, disabled=False,
                                             layout=ipywidgets.Layout(width='250px'))
        color_input.observe(on_color_change, names='value')

        controls_vbox = ipywidgets.VBox([toggle_input, name_input, color_input])

        return controls_vbox

    def __prepare_surface_controls(self, actor):
        surface_type = ipywidgets.ToggleButtons(options=['Surface', 'Wireframe', 'Points'], disabled=False,
                                                layout=ipywidgets.Layout(margin='0px 0px 0px 25px'))
        surface_type.style.button_width = '70px'

        def toggle_cortex_type(change):
            if change['new'] == 'Wireframe':
                actor.GetProperty().SetRepresentationToWireframe()
            elif change['new'] == 'Surface':
                actor.GetProperty().SetRepresentationToSurface()
            else:
                actor.GetProperty().SetRepresentationToPoints()
            self.output_plot.update_plot()

        surface_type.observe(toggle_cortex_type, 'value')

        surface_opacity = ipywidgets.FloatSlider(value=1, min=0, max=1.0, step=0.1, description='Opacity:',
                                                 disabled=False, continuous_update=False, orientation='horizontal',
                                                 readout=True, readout_format='.1f',
                                                 layout=ipywidgets.Layout(width='330px'))

        def on_opacity_change(change):
            value = change['new']
            actor.GetProperty().SetOpacity(value)
            self.output_plot.update_plot()

        surface_opacity.observe(on_opacity_change, names='value')

        return surface_opacity, surface_type

    def __prepare_points_controls(self, actor, config):
        def on_size_change(change):
            value = change['new']
            actor.GetProperty().SetPointSize(value)
            self.output_plot.update_plot()

        size_input = ipywidgets.IntText(value=config.size, description='Size: ', disabled=False,
                                        layout=ipywidgets.Layout(width='250px'))
        size_input.observe(on_size_change, names='value')

        return size_input,


class HeadBrowser(ipywidgets.VBox, TVBWidgetWithBrowser):

    def __init__(self):
        super().__init__()
        surface_button = ipywidgets.Button(description='View surface')
        sensors_button = ipywidgets.Button(description='View sensors')
        connectivity_button = ipywidgets.Button(description='View connectivity')
        self.buttons = ipywidgets.HBox([surface_button, sensors_button, connectivity_button],
                                       layout=ipywidgets.Layout(margin="0px 0px 0px 20px"))
        self.head_widget = HeadWidget()
        self.children = [self.storage_widget, self.buttons, self.message_label, self.head_widget]

        def add_surface_datatype(_):
            self.load_selected_file(Surface)

        def add_sensors_datatype(_):
            self.load_selected_file(Sensors, ('.txt', '.txt.bz2'))

        def add_connectivity_datatype(_):
            self.load_selected_file(Connectivity)

        surface_button.on_click(add_surface_datatype)
        sensors_button.on_click(add_sensors_datatype)
        connectivity_button.on_click(add_connectivity_datatype)

    def add_datatype(self, datatype, config=None):
        # type: (HasTraits, HeadWidgetConfig) -> None
        self.head_widget.add_datatype(datatype, config)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os
import math
import mne
import numpy as np
import ipywidgets as widgets
import matplotlib.pyplot as plt
from abc import ABC, abstractmethod
from IPython.core.display_functions import display
from tvb.datatypes.time_series import TimeSeries
from tvbwidgets.core.ini_parser import parse_ini_file
from tvbwidgets.core.exceptions import InvalidInputException
from tvbwidgets.ui.base_widget import TVBWidget
from tvbwidgets.ui.widget_with_browser import TVBWidgetWithBrowser

mne.set_config('MNE_BROWSER_BACKEND', 'matplotlib')


class ABCDataWrapper(ABC):
    """ Wrap any TimeSeries for TSWidget to read/parse uniformly"""
    extra_dimensions = {1: ("State var.", None),
                        3: ("Mode", None)}
    CHANNEL_TYPE = "misc"
    MAX_DISPLAYED_TIMEPOINTS = 3000

    @property
    def data_shape(self):
        # type: () -> tuple
        return ()

    @property
    def displayed_time_points(self):
        # type: () -> int
        return min(self.data_shape[0], self.MAX_DISPLAYED_TIMEPOINTS)

    @abstractmethod
    def get_channels_info(self):
        # type: () -> (list, list, list)
        pass

    @abstractmethod
    def get_ts_period(self):
        # type: () -> float
        pass

    @abstractmethod
    def get_ts_sample_rate(self):
        # type: () -> float
        pass

    @abstractmethod
    def build_raw(self, np_slice=None):
        # type: (tuple) -> mne.io.RawArray
        pass

    @abstractmethod
    def get_update_slice(self, sel1, sel2):
        # type: (int, int) -> tuple
        pass

    @abstractmethod
    def get_slice_for_time_point(self, time_point, channel, sel1=0, sel2=0):
        # type: (int, int, int, int) -> tuple
        pass

    @abstractmethod
    def get_hover_channel_value(self, x, ch_index, sel1, sel2):
        # type: (float, int, int, int) -> float
        pass


class WrapperTVB(ABCDataWrapper):
    """ Wrap TVB TimeSeries object for tsWidget"""

    def __init__(self, data):
        # type: (TimeSeries) -> None
        if data is None or not isinstance(data, TimeSeries):
            raise InvalidInputException("Not a valid TVB TS " + str(data))
        self.data = data
        self.ch_names = []
        variables_labels = data.variables_labels
        if variables_labels is not None and variables_labels != []:
            sv_options = [(variables_labels[idx], idx) for idx in range(len(variables_labels))]
            self.extra_dimensions = ABCDataWrapper.extra_dimensions.copy()
            self.extra_dimensions[1] = ("State var.", sv_options)

    @property
    def data_shape(self):
        # type: () -> tuple
        return self.data.shape

    def get_channels_info(self):
        # type: () -> (list, list, list)
        no_channels = self.data.shape[2]  # number of channels is on axis 2

        if hasattr(self.data, "connectivity"):
            ch_names = self.data.connectivity.region_labels.tolist()
        elif hasattr(self.data, "sensors"):
            ch_names = self.data.sensors.labels.tolist()
        else:
            ch_names = ['signal-%d' % i for i in range(no_channels)]

        ch_order = list(range(no_channels))  # the order should be the order in which they are provided
        ch_types = [self.CHANNEL_TYPE for _ in ch_names]
        self.ch_names = ch_names
        return ch_names, ch_order, ch_types

    def get_ts_period(self):
        # type: () -> float
        displayed_period = self.data.sample_period * self.displayed_time_points
        return displayed_period

    def get_ts_sample_rate(self):
        # type: () -> float
        return self.data.sample_rate

    def build_raw(self, np_slice=None):
        # type: (tuple) -> mne.io.RawArray
        if np_slice is None:
            np_slice = self.get_update_slice()
        raw_info = mne.create_info(self.ch_names, sfreq=self.data.sample_rate)
        data_for_raw = self.data.data[np_slice].squeeze()
        data_for_raw = np.swapaxes(data_for_raw, 0, 1)
        raw = mne.io.RawArray(data_for_raw, raw_info, first_samp=self.data.start_time * self.data.sample_rate)
        return raw

    def get_update_slice(self, sel1=0, sel2=0):
        # type: (int, int) -> tuple
        new_slice = (slice(None), slice(sel1, sel1 + 1), slice(None), slice(sel2, sel2 + 1))
        return new_slice

    def get_slice_for_time_point(self, time_point, channel, sel1=0, sel2=0):
        # type: (int, int, int, int) -> tuple
        new_slice = (slice(time_point, time_point + 1), slice(sel1, sel1 + 1),
                     slice(channel, channel + 1), slice(sel2, sel2 + 1))
        return new_slice

    def get_hover_channel_value(self, x, ch_index, sel1, sel2):
        # type: (float, int, int, int) -> float
        time_per_tp = self.get_ts_period() / self.displayed_time_points  # time unit displayed in 1 time point

        # which time point to search in the time points array
        tp_on_hover = round((x - self.data.start_time) / time_per_tp)
        new_slice = self.get_slice_for_time_point(tp_on_hover, ch_index, sel1, sel2)

        ch_value = self.data.data[new_slice].squeeze().item(0)
        ch_value = round(ch_value, 4)

        return ch_value


class WrapperNumpy(ABCDataWrapper):
    """ Wrap a numpy array for tsWidget """

    def __init__(self, data, sample_rate, channel_names=None, ch_idx=2):
        # type: (np.ndarray, float, list, int) -> None
        """
        :param data: Numpy array 2D up to 4D
        :param sample_rate: float
        :param channel_names: optional names for channels
        :param ch_idx: Channels Index from the max 4D of the data. We assume time is on 0
        """
        if data is None or not isinstance(data, np.ndarray) or len(data.shape) <= 1:
            raise InvalidInputException("Not a valid numpy array %s \n "
                                        "It should be numpy.ndarray, at least 2D up to 4D" % str(data))
        self.data = data
        self.sample_rate = sample_rate
        self.ch_names = channel_names or []
        self.ch_idx = ch_idx

    @property
    def data_shape(self):
        # type: () -> tuple
        return self.data.shape

    def get_channels_info(self):
        # type: () -> (list, list, list)
        no_channels = self.data.shape[self.ch_idx]
        if (self.ch_names is None) or len(self.ch_names) != no_channels:
            self.ch_names = ['signal-%d' % i for i in range(no_channels)]
        ch_order = list(range(no_channels))  # the order should be the order in which they are provided
        ch_types = [self.CHANNEL_TYPE for _ in self.ch_names]
        return self.ch_names, ch_order, ch_types

    def get_ts_period(self):
        # type: () -> float
        sample_period = 1 / self.sample_rate
        displayed_period = sample_period * self.displayed_time_points
        return displayed_period

    def get_ts_sample_rate(self):
        # type: () -> float
        return self.sample_rate

    def build_raw(self, np_slice=None):
        # type: (tuple) -> mne.io.RawArray
        if np_slice is None:
            np_slice = self.get_update_slice()
        raw_info = mne.create_info(self.ch_names, sfreq=self.sample_rate)
        data_for_raw = self.data[np_slice]
        data_for_raw = data_for_raw.squeeze()
        data_for_raw = np.swapaxes(data_for_raw, 0, 1)
        raw = mne.io.RawArray(data_for_raw, raw_info)
        return raw

    def get_update_slice(self, sel1=0, sel2=0):
        # type: (int, int) -> tuple
        sel2 = sel2 if sel2 is not None else 0
        no_dim = len(self.data_shape)
        dim_to_slice_dict = {
            2: (slice(None), slice(None)),
            3: (slice(None), slice(sel1, sel1 + 1), slice(None)),
            4: (slice(None), slice(sel1, sel1 + 1), slice(None), slice(sel2, sel2 + 1))
        }
        new_slice = dim_to_slice_dict[no_dim]
        return new_slice

    def get_slice_for_time_point(self, time_point, channel, sel1=0, sel2=0):
        # type: (int, int, int, int) -> tuple
        sel1 = sel1 if sel1 is not None else 0
        sel2 = sel2 if sel2 is not None else 0
        no_dim = len(self.data_shape)
        dim_to_slice_dict = {
            2: (slice(time_point, time_point + 1), slice(channel, channel + 1)),
            3: (slice(time_point, time_point + 1), slice(sel1, sel1 + 1), slice(channel, channel + 1)),
            4: (slice(time_point, time_point + 1), slice(sel1, sel1 + 1),
                slice(channel, channel + 1), slice(sel2, sel2 + 1))
        }
        new_slice = dim_to_slice_dict[no_dim]
        return new_slice

    def get_hover_channel_value(self, x, ch_index, sel1, sel2):
        # type: (float, int, int, int) -> float
        time_per_tp = self.get_ts_period() / self.displayed_time_points  # time unit displayed in 1 time point

        # which time point to search in the time points array
        tp_on_hover = round(x / time_per_tp)
        new_slice = self.get_slice_for_time_point(tp_on_hover, ch_index, sel1, sel2)

        ch_value = self.data[new_slice].squeeze().item(0)
        ch_value = round(ch_value, 4)

        return ch_value


class TimeSeriesWidget(widgets.VBox, TVBWidget):
    """ Actual TimeSeries Widget """

    def __init__(self, **kwargs):

        self.fig = None
        self.data = None
        self.ch_names = []
        self.ch_order = []
        self.ch_types = []
        self.displayed_period = 0
        self.no_channels = 30
        self.raw = None
        self.sample_freq = 0

        self.output = widgets.Output(layout=widgets.Layout(width='auto'))
        annotation_area = self._create_annotation_area()
        self.instr_area = self._create_instructions_region()
        self.title_area = widgets.HBox(children=[self.instr_area])

        self.checkboxes = dict()
        super().__init__([self.output, annotation_area, self.title_area], layout=self.DEFAULT_BORDER)
        self.logger.info("TimeSeries Widget initialized")

    def add_datatype(self, ts_tvb):
        # type: (TimeSeries) -> None
        data_wrapper = WrapperTVB(ts_tvb)
        self.logger.debug("Adding TVB TS for display...")
        self._populate_from_data_wrapper(data_wrapper)

    def reset_data(self):
        self.data = None
        self.title_area.children = [self.instr_area]

    def _populate_from_data_wrapper(self, data_wrapper):
        # type: (ABCDataWrapper) -> None
        if self.data is not None:
            raise InvalidInputException("TSWidget is not yet capable to display more than one TS, "
                                        "either use wid.reset_data, or create another widget instance!")

        self.data = data_wrapper
        self.sample_freq = data_wrapper.get_ts_sample_rate()
        self.displayed_period = data_wrapper.get_ts_period()
        self.ch_names, self.ch_order, self.ch_types = data_wrapper.get_channels_info()
        self.raw = self.data.build_raw()
        channels_area = self._create_checkboxes(data_wrapper)
        self.title_area.children += (channels_area,)
        self._redraw()

    def add_data_array(self, numpy_array, sample_freq, ch_idx):
        # type: (np.array, float, int) -> None
        data_wrapper = WrapperNumpy(numpy_array, sample_freq, ch_idx=ch_idx)
        self._populate_from_data_wrapper(data_wrapper)

    # ======================================== CHANNEL VALUE AREA ======================================================
    def _create_annotation_area(self):
        title_label = widgets.Label(value='Channel values:')
        self.channel_val_area = widgets.VBox()
        annot_area = widgets.HBox(children=[title_label, self.channel_val_area], layout={'height': '70px',
                                                                                         'padding': '0 0 0 100px'})
        return annot_area

    # ===================================== INSTRUCTIONS DROPDOWN ======================================================
    @staticmethod
    def _create_instructions_region():
        instr_list, key_list, val_list = [], [], []
        help_text = parse_ini_file(os.path.join(os.path.dirname(__file__), "ts_widget_help.ini"))

        for key, value in help_text.items():
            key_label = widgets.Label(value=key)
            val_label = widgets.Label(value=value)
            key_list.append(key_label)
            val_list.append(val_label)

        instr_list.append(widgets.VBox(children=key_list))
        instr_list.append(widgets.VBox(children=val_list))
        instr_region = widgets.HBox(children=instr_list)
        instr_accordion = widgets.Accordion(children=[instr_region], selected_index=None,
                                            layout=widgets.Layout(width='40%'))
        instr_accordion.set_title(0, 'Keyboard shortcuts')
        return instr_accordion

    # =========================================== PLOT =================================================================
    def _redraw(self):
        def update_on_plot_interaction(_):
            """
            Function that updates the checkboxes when the user navigates through the plot
            using either the mouse or the keyboard
            """
            picks = list(self.fig.mne.picks)
            for ch in picks:
                ch_name = self.fig.mne.ch_names[ch]
                cb = self.checkboxes[ch_name]
                if not cb.value:
                    cb.value = True
            self._update_fig()

        def hover(event):
            self.channel_val_area.children = []

            values = []  # list of label values for channels we are hovering over
            x = event.xdata  # time unit (s, ms) we are hovering over
            lines = self.fig.mne.traces  # all currently visible channels
            sel1, sel2 = self._get_selection_values()
            if event.inaxes == self.fig.mne.ax_main:
                for line in lines:
                    if line.contains(event)[0]:
                        line_index = lines.index(line)  # channel index among displayed channels
                        ch_index = self.fig.mne.picks[line_index]  # channel index among all channels
                        ch_name = self.fig.mne.ch_names[ch_index]

                        ch_value = self.data.get_hover_channel_value(x, ch_index, sel1, sel2)
                        label_val = f'{ch_name}: {ch_value}'
                        values.append(label_val)
                for v in values:
                    val_label = widgets.Label(value=v)
                    self.channel_val_area.children += (val_label,)

        # display the plot
        with plt.ioff():
            # create the plot
            self.fig = self.raw.plot(duration=self.displayed_period,
                                     n_channels=self.no_channels, show_first_samp=True,
                                     clipping=None, show=False)
            self.fig.set_size_inches(11, 5)
            # self.fig.figure.canvas.set_window_title('TimeSeries plot')
            self.fig.mne.ch_order = self.ch_order
            self.fig.mne.ch_types = np.array(self.ch_types)

            # add custom widget handling on keyboard and mouse events
            self.fig.canvas.mpl_connect('key_press_event', update_on_plot_interaction)
            self.fig.canvas.mpl_connect('button_press_event', update_on_plot_interaction)
            self.fig.canvas.mpl_connect("motion_notify_event", hover)
            with self.output:
                self.output.clear_output(wait=True)
                display(self.fig.canvas)

    # ======================================== CHANNELS  ==============================================================
    def _unselect_all(self, _):
        self.logger.debug("Unselect all was called!")
        for cb_name in self.checkboxes:
            self.checkboxes[cb_name].value = False

    def _select_all(self, _):
        self.logger.debug("Select all was called!")
        for cb_name in self.checkboxes:
            self.checkboxes[cb_name].value = True

    def _create_checkboxes(self, array_wrapper):
        # type: (ABCDataWrapper) -> widgets.Accordion
        checkboxes_list, checkboxes_stack = [], []
        labels = array_wrapper.get_channels_info()[0]
        cb_per_col = math.ceil(len(labels) / 7)  # number of checkboxes in a column; should always display 7 cols
        for i, label in enumerate(labels):
            self.checkboxes[label] = widgets.Checkbox(value=True, description=label,
                                                      disabled=False, indent=False)
            self.checkboxes[label].observe(self._update_ts, names="value", type="change")
            if i and i % cb_per_col == 0:
                checkboxes_list.append(widgets.VBox(children=checkboxes_stack))
                checkboxes_stack = []
            checkboxes_stack.append(self.checkboxes[label])
        checkboxes_list.append(widgets.VBox(children=checkboxes_stack))
        checkboxes_region = widgets.HBox(children=checkboxes_list, layout={'width': '540px',
                                                                           'height': 'max-content'})

        # buttons region
        select_all_btn = widgets.Button(description="Select all", layout=self.BUTTON_STYLE)
        select_all_btn.on_click(self._select_all)
        unselect_all_btn = widgets.Button(description="Unselect all", layout=self.BUTTON_STYLE)
        unselect_all_btn.on_click(self._unselect_all)
        actions = [select_all_btn, unselect_all_btn]

        # select dimensions region
        self.radio_buttons = []
        for idx, info in array_wrapper.extra_dimensions.items():
            extra_area, extra_radio_btn = self._create_selection(info[0], idx, dim_options=info[1])
            self.radio_buttons.append(extra_radio_btn)
            if extra_area is not None:
                actions.append(extra_area)

        channels_region = widgets.VBox(children=[widgets.HBox(actions), checkboxes_region])
        channels_area = widgets.Accordion(children=[channels_region], selected_index=None,
                                          layout=widgets.Layout(width='50%'))
        channels_area.set_title(0, 'Channels')
        return channels_area

    def _create_selection(self, title="Mode", shape_pos=3, dim_options=None):

        if self.data is None or len(self.data.data_shape) <= max(2, shape_pos):
            return None, None

        no_dims = self.data.data_shape[shape_pos]
        if dim_options is None or dim_options == []:
            dim_options = [i for i in range(no_dims)]
        sel_radio_btn = widgets.RadioButtons(options=dim_options, layout={'width': 'max-content'})
        sel_radio_btn.observe(self._dimensions_selection_update, names=['value'])
        accordion = widgets.Accordion(children=[sel_radio_btn], selected_index=None, layout={'width': '30%'})
        accordion.set_title(0, title)
        return accordion, sel_radio_btn

    def _get_selection_values(self):
        sel1 = self.radio_buttons[0].value if self.radio_buttons[0] else None
        sel2 = self.radio_buttons[1].value if self.radio_buttons[1] else None
        return sel1, sel2

    def _dimensions_selection_update(self, _):
        # update self.raw and linked parts
        sel1, sel2 = self._get_selection_values()
        new_slice = self.data.get_update_slice(sel1, sel2)
        self.logger.info("New slice " + str(new_slice))
        self.raw = self.data.build_raw(new_slice)
        self.fig = self.raw.plot(duration=self.displayed_period,
                                 n_channels=self.no_channels,
                                 clipping=None, show=False)
        self.fig.set_size_inches(11, 5)

        self._redraw()

        # refresh the checkboxes if they were unselected
        for cb_name in self.checkboxes:
            self.checkboxes[cb_name].value = True

    def _update_ts(self, val):
        ch_names = list(self.fig.mne.ch_names)
        self.logger.debug("Update_ts is called for channels " + str(ch_names))

        # check if newly checked option is before current ch_start in the channels list
        if (val['old'] is False) and (val['new'] is True):
            ch_name = val['owner'].description
            ch_number = ch_names.index(ch_name)
            ch_changed_index = list(self.fig.mne.ch_order).index(ch_number)
            if ch_changed_index < self.fig.mne.ch_start:
                self.fig.mne.ch_start = ch_changed_index

        # divide list of all channels into checked(picked) and unchecked(not_picked) channels
        picks = []
        not_picked = []
        for cb in list(self.checkboxes.values()):
            ch_number = ch_names.index(cb.description)  # get the number representation of checked/unchecked channel
            if cb.value:
                picks.append(ch_number)  # list with number representation of channels
            else:
                not_picked.append(ch_number)

        # for unselect all
        if not picks:
            self.fig.mne.picks = picks
            self.fig.mne.n_channels = 0
            self._update_fig()
            return

        # if not enough values are checked, force the plot to display less channels
        n_channels = self.fig.mne.n_channels
        if (len(picks) < n_channels) or (n_channels < len(picks) <= self.no_channels):
            self.fig.mne.n_channels = len(picks)
        else:
            self.fig.mne.n_channels = self.no_channels

        # order list of checked channels according to ordering rule (self.fig.mne.ch_order)
        ch_order_filtered = [x for x in self.fig.mne.ch_order if x not in not_picked]

        ch_start = self.fig.mne.ch_start
        ch_start_number = self.fig.mne.ch_order[ch_start]
        if ch_start_number not in ch_order_filtered:  # if first channel was unchecked
            ch_start = self._get_next_checked_channel(ch_start, ch_order_filtered, list(self.fig.mne.ch_order))

        self.fig.mne.ch_start = ch_start
        ch_start_number = self.fig.mne.ch_order[self.fig.mne.ch_start]
        ch_start_index = ch_order_filtered.index(ch_start_number)

        new_picks = np.array(ch_order_filtered[ch_start_index:(ch_start_index + self.fig.mne.n_channels)])
        self.fig.mne.n_channels = len(new_picks)  # needed for WID-66
        self.fig.mne.picks = new_picks
        ch_start_index = list(self.fig.mne.ch_order).index(new_picks[0])
        self.fig.mne.ch_start = ch_start_index

        self._update_fig()

    @staticmethod
    def _get_next_checked_channel(ch_start, checked_list, ch_order):
        for _ in range(len(ch_order)):
            ch_start += 1
            new_ch_start_number = ch_order[ch_start]  # get the number representation of next first channel
            if new_ch_start_number in checked_list:
                break
        return ch_start

    def _update_fig(self):
        self.fig._update_trace_offsets()
        self.fig._update_vscroll()
        try:
            self.fig._redraw(annotations=True)
        except:
            self.fig._redraw(update_data=False)  # needed in case of Unselect all


class TimeSeriesBrowser(widgets.VBox, TVBWidgetWithBrowser):

    def __init__(self, collab=None, folder=None):
        super().__init__(**{'collab': collab, 'folder': folder})
        timeseries_button = widgets.Button(description='View time series')
        self.buttons = widgets.HBox([timeseries_button], layout=widgets.Layout(margin="0px 0px 0px 20px"))
        self.timeseries_widget = TimeSeriesWidget()
        self.children = [self.storage_widget, self.buttons, self.message_label, self.timeseries_widget]

        def add_timeseries_datatype(_):
            self.load_selected_file(TimeSeries, ('.h5', '.npz'))

        timeseries_button.on_click(add_timeseries_datatype)

    def add_datatype(self, datatype):  # type: (TimeSeries) -> None
        self.timeseries_widget.reset_data()
        self.timeseries_widget.add_datatype(datatype)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import pytest
import logging
import numpy as np
from ipywidgets import Checkbox
from tvbwidgets.tests.ts_generator import generate_ts_with_mode_and_sv
from tvbwidgets.ui.ts_widget import TimeSeriesWidget, WrapperTVB, WrapperNumpy


# =========================================== TEST WRAPPER NUMPY =======================================================
@pytest.fixture
def wrapper_np():
    """ Returns an initialized Numpy wrapper with 3 dimensions """
    numpy_array = np.random.rand(30000, 4, 50)
    wrapper_np = WrapperNumpy(numpy_array, 0.01, ch_idx=2)
    return wrapper_np


def test_data_shape_wrapper_np(wrapper_np):
    assert wrapper_np.data_shape == (30000, 4, 50)


def test_get_channel_info_wrapper_np(wrapper_np):
    ch_names, ch_order, ch_type = wrapper_np.get_channels_info()
    assert len(ch_names) == len(ch_order) == len(ch_type) == 50


def test_get_ts_period_wrapper_np(wrapper_np):
    assert wrapper_np.get_ts_period() == 300000


def test_get_sample_rate_wrapper_np(wrapper_np):
    assert wrapper_np.get_ts_sample_rate() == 0.01


def test_build_raw_wrapper_np(wrapper_np):
    wrapper_np.get_channels_info()  # need to init ch_names for wrapper

    raw = wrapper_np.build_raw()
    data = wrapper_np.data[:, 0, :].squeeze()
    data = np.swapaxes(data, 0, 1)

    assert np.array_equal(raw.get_data(), data)


def test_get_update_slice_wrapper_np(wrapper_np):
    sel1 = 1
    new_slice = wrapper_np.get_update_slice(sel1)
    data_with_slice = wrapper_np.data[new_slice].squeeze()
    new_data = wrapper_np.data[:, sel1, :]
    assert np.array_equal(data_with_slice, new_data)


def test_get_hover_channel_value_wrapper_np(wrapper_np):
    time_per_tp = 100
    x = 2999900
    x_int = round(x / time_per_tp)
    ch_index = 0
    sel1 = 0

    ch_value = wrapper_np.get_hover_channel_value(x, ch_index, sel1, None)
    val = wrapper_np.data[x_int, sel1, ch_index]
    val = round(val, 4)

    assert ch_value == val


# ============================================ TEST WRAPPER TVB ========================================================
@pytest.fixture(scope="module")
def tsr_4d():
    """ Returns a TVB TS having SV and Mode """
    return generate_ts_with_mode_and_sv(length=1e3, cutoff=100)


@pytest.fixture
def wrapper_tvb(tsr_4d):
    """ Returns an initialized WrapperTVB """
    return WrapperTVB(tsr_4d)


def test_build_wrapper_tvb(wrapper_tvb):
    assert wrapper_tvb.data_shape == (4000, 4, 76, 1)

    ch_names, ch_order, ch_type = wrapper_tvb.get_channels_info()
    assert len(ch_names) == len(ch_order) == len(ch_type) == 76

    assert wrapper_tvb.displayed_time_points == 3000
    assert wrapper_tvb.get_ts_period() == 0.75
    assert wrapper_tvb.get_ts_sample_rate() == 4000

    raw = wrapper_tvb.build_raw()
    data = wrapper_tvb.data.data[:, 0, :, 0].squeeze()
    data = np.swapaxes(data, 0, 1)

    assert np.array_equal(raw.get_data(), data)


def test_get_update_slice_wrapper_tvb(wrapper_tvb):
    sel1 = 2
    sel2 = 0
    new_slice = wrapper_tvb.get_update_slice(sel1, sel2)
    data_with_slice = wrapper_tvb.data.data[new_slice].squeeze()
    new_data = wrapper_tvb.data.data[:, sel1, :, sel2]

    assert np.array_equal(data_with_slice, new_data)


def test_get_hover_channel_value_wrapper_tvb(wrapper_tvb):
    start_time = 0.100125
    time_per_tp = 0.00025

    x = 1.099
    x_int = round((x - start_time) / time_per_tp)
    ch_index = 6
    sel1 = 0
    sel2 = 0

    ch_value = wrapper_tvb.get_hover_channel_value(x, ch_index, sel1, sel2)
    val = wrapper_tvb.data.data[x_int, sel1, ch_index, sel2]
    val = round(val, 4)
    assert ch_value == val


# ============================================= TEST TS WIDGET =========================================================
@pytest.fixture
def tsw():
    """ Returns an empty TS widget """
    return TimeSeriesWidget()


@pytest.fixture
def tsw_tvb_data(tsr_4d):
    """ Returns a TS widget initialized with a tvb data wrapper"""
    tsw_tvb_data = TimeSeriesWidget()
    tsw_tvb_data.add_datatype(tsr_4d)
    return tsw_tvb_data


def test_create_ts_widget(tsw):
    assert tsw.ch_names == []
    assert tsw.ch_order == []
    assert tsw.ch_types == []
    assert tsw.displayed_period == 0
    assert tsw.no_channels == 30
    assert tsw.sample_freq == 0


def test_populate_from_data_wrapper_tvb(tsw, wrapper_tvb):
    tsw._populate_from_data_wrapper(wrapper_tvb)

    assert tsw.data == wrapper_tvb
    assert tsw.sample_freq == 4000
    assert tsw.displayed_period == 0.75
    assert len(tsw.ch_names) == len(tsw.ch_order) == len(tsw.ch_types) == 76
    assert tsw.raw.get_data().shape == (76, 4000)


def test_create_selection(tsw_tvb_data):
    title = "State variable"
    pos = 1
    accordion, radio_button = tsw_tvb_data._create_selection(title=title, shape_pos=pos)
    assert radio_button.options == (0, 1, 2, 3)
    assert accordion.get_title(0) == title


def test_create_checkboxes(tsw_tvb_data):
    tsw_tvb_data.data.get_channels_info()  # need to initialize ch_names for wrapper
    channels_area = tsw_tvb_data._create_checkboxes(tsw_tvb_data.data)
    assert len(tsw_tvb_data.checkboxes) == 76
    assert len(tsw_tvb_data.radio_buttons) == 2
    assert channels_area.get_title(0) == 'Channels'


def test_get_widget(tsw_tvb_data):
    tsw_tvb_data.get_widget()
    assert type(tsw_tvb_data) == TimeSeriesWidget


def test_update_ts(caplog, tsw_tvb_data):
    logger = logging.getLogger('tvbwidgets')
    logger.propagate = True
    val = {'name': 'value', 'old': True, 'new': False,
           'owner': Checkbox(value=False, description='signal-0', indent=False), 'type': 'change'}

    caplog.clear()
    tsw_tvb_data.get_widget()
    tsw_tvb_data._update_ts(val)
    assert caplog.records[0].levelname == 'DEBUG'
    assert 'Update_ts' in caplog.text


def test_create_instructions_region():
    instr_area = TimeSeriesWidget._create_instructions_region()

    shortcuts = instr_area.children[0].children[0].children
    descriptions = instr_area.children[0].children[1].children

    no_instruction = 23
    assert len(shortcuts) == len(descriptions) == no_instruction

    title = 'Keyboard shortcuts'
    assert instr_area.get_title(0) == title


def test_get_next_checked_channel():
    ch_start = 0
    checked_list = [2, 3, 4, 5]
    ch_order = [0, 1, 2, 3, 4, 5]
    assert TimeSeriesWidget._get_next_checked_channel(ch_start, checked_list, ch_order) == 2



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

from tvbwidgets.core.simulator.tvb_integrators import IntegratorsEnum


def test_get_integrators_dict():
    integrators = IntegratorsEnum.get_integrators_dict()
    assert len(integrators.keys()) == 15



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

from tvb.simulator.lab import *
from tvb.datatypes import time_series
import tvb.datatypes.time_series
import numpy as np
import math


def generate_ts_with_stimulus(length=5e3, cutoff=1e3, conn=None):
    if conn is None:
        conn = _generate_connectivity(76)

    weighting = np.zeros((76,))
    weighting[[14, 52, 11, 49]] = 0.1

    eqn_t = equations.PulseTrain()
    eqn_t.parameters['onset'] = 1.5e3
    eqn_t.parameters['T'] = 100.0
    eqn_t.parameters['tau'] = 50.0

    stimulus = patterns.StimuliRegion(
        temporal=eqn_t,
        connectivity=conn,
        weight=weighting)
    stimulus.configure_space()
    stimulus.configure_time(np.arange(0., 3e3, 2 ** -4))

    sim = simulator.Simulator(model=models.Generic2dOscillator(a=np.array([0.3]), tau=np.array([2])),
                              connectivity=conn,
                              coupling=coupling.Difference(a=np.array([7e-4])),
                              integrator=integrators.HeunStochastic(dt=0.5,
                                                                    noise=noise.Additive(nsig=np.array([5e-5]))),
                              monitors=(monitors.TemporalAverage(period=1.0),),
                              stimulus=stimulus,
                              simulation_length=length + cutoff,
                              ).configure()

    (tavg_time, tavg_data), = sim.run()
    cut_idx = int(math.ceil(cutoff / sim.monitors[0].period))

    tsr = tvb.datatypes.time_series.TimeSeriesRegion(data=tavg_data[cut_idx:],
                                                     time=tavg_time[cut_idx:],
                                                     connectivity=conn,
                                                     start_time=tavg_time[cut_idx] / 1000.0,
                                                     sample_period=sim.monitors[0].period / 1000.0,
                                                     sample_period_unit="s")
    # Fill the State Var names
    state_variable_dimension_name = tsr.labels_ordering[1]
    selected_vois = [sim.model.variables_of_interest[idx] for idx in sim.monitors[0].voi]
    tsr.labels_dimensions[state_variable_dimension_name] = selected_vois

    tsr.configure()
    return tsr


def generate_ts_with_mode_and_sv(length=5e3, cutoff=500, conn=None):
    if conn is None:
        conn = _generate_connectivity(76)
    jrm = models.JansenRit(mu=np.array([0.]), v0=np.array([6.]))
    monitor = monitors.TemporalAverage(period=2 ** -2)

    phi_n_scaling = (jrm.a * jrm.A * (jrm.p_max - jrm.p_min) * 0.5) ** 2 / 2.
    sigma = np.zeros(6)
    sigma[3] = phi_n_scaling

    sim = simulator.Simulator(model=jrm,
                              connectivity=conn,
                              coupling=coupling.SigmoidalJansenRit(a=np.array([10.0])),
                              integrator=integrators.HeunStochastic(dt=2 ** -4, noise=noise.Additive(nsig=sigma)),
                              monitors=(monitor,),
                              simulation_length=length + cutoff,
                              ).configure()

    (time_array, data_array), = sim.run()
    cut_idx = int(math.ceil(cutoff / monitor.period))

    tsr = time_series.TimeSeriesRegion(data=data_array[cut_idx:],
                                       time=time_array[cut_idx:],
                                       connectivity=conn,
                                       start_time=time_array[cut_idx] / 1000.0,
                                       sample_period=monitor.period / 1000,
                                       sample_period_unit="s")
    # Fill the State Var names
    state_variable_dimension_name = tsr.labels_ordering[1]
    selected_vois = [jrm.variables_of_interest[idx] for idx in monitor.voi]
    tsr.labels_dimensions[state_variable_dimension_name] = selected_vois

    tsr.configure()
    return tsr


def _generate_connectivity(no_of_regions):
    labels = np.array(['sig ' + str(i) for i in range(no_of_regions)])
    conn = connectivity.Connectivity(centres=np.random.rand(no_of_regions, 3),
                                     region_labels=labels,
                                     weights=np.random.rand(no_of_regions, no_of_regions),
                                     tract_lengths=np.random.rand(no_of_regions, no_of_regions))
    return conn



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

class WidgetsException(Exception):

    def __init__(self, message):
        super().__init__(message)
        self.message = str(message)

    def __str__(self):
        return self.message


class InvalidInputException(WidgetsException):
    """
    To be thrown when an invalid input is received.
    """


class InvalidFileException(WidgetsException):
    """
    To be thrown if an invalid filename is used.
    """


class ModelNotFoundError(WidgetsException):
    """
    To be thrown when a model class can't be found or doesn't exist
    """


class ModelExporterNotFoundError(WidgetsException):
    """
    To be thrown when an attempt is being made to use an exporter that is not accessible
    """



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os

import pytest
from ebrains_drive.files import SeafFile, SeafDir
from ebrains_drive.repo import Repo

from tvbwidgets.core.auth import CLB_AUTH
from tvbwidgets.ui.drive_widget import DriveWidget
from tvbwidgets.ui.storage_widget import StorageWidget

DUMMY_FILENAME = '/dummy_file.txt'
DUMMY_DIR = '/dummy_dir'
DUMMY_CONTENT = "Dummy content"


class MockSeafFile(SeafFile):
    def get_content(self):
        return DUMMY_CONTENT


class MockSeafDir(SeafDir):
    def __init__(self, repo, path, object_id, obj_type, empty_dir=False):
        super().__init__(repo, path, object_id, obj_type)
        self.empty_dir = empty_dir

    def ls(self, entity_type=None, force_refresh=True):
        if self.empty_dir:
            return list()
        return [MockSeafFile(self.repo, DUMMY_FILENAME, 1, SeafFile),
                MockSeafDir(self.repo, DUMMY_DIR, 2, SeafDir, empty_dir=True)]


class MockRepo(Repo):

    def __init__(self, client, **kwargs):
        super(MockRepo, self).__init__(client, **kwargs)
        self.dir = MockSeafDir(self, '/', 1, SeafDir)

    def get_dir(self, path):
        if path == DriveWidget.ROOT:
            return self.dir
        return self.dir.ls()[1]

    def get_file(self, path):
        if path is None:
            return None
        return self.dir.ls()[0]


class MockRepos:

    def __init__(self, client):
        self.client = client

    def list_repos(self):
        return [MockRepo(self.client, name='repo1', id=1), MockRepo(self.client, name='repo2', id=2)]


class MockDriveClient:
    @property
    def repos(self):
        return MockRepos(self)


def test_drive_widget(mocker):
    def mockk(token):
        return MockDriveClient()

    mocker.patch('ebrains_drive.connect', mockk)

    if os.environ.get(CLB_AUTH):
        os.environ.pop(CLB_AUTH)

    with pytest.raises(RuntimeError):
        DriveWidget()

    os.environ[CLB_AUTH] = "test_auth_token"
    widget = DriveWidget()

    assert widget.get_chosen_repo() is None

    # this triggers the observe events on dropbox widget
    widget.repos_dropdown.value = widget.repos_dropdown.options['repo1']
    assert isinstance(widget.get_chosen_repo(), Repo)
    assert widget.get_selected_file_path() is None
    assert len(widget.files_list.options) == 3  # 3 because the parent dir is also added

    # this triggers the observe events on select widget
    widget.files_list.value = widget.files_list.options[1]
    assert widget.get_selected_file_path() == DUMMY_FILENAME
    assert widget.get_selected_file_content() == DUMMY_CONTENT

    # this triggers the observe events on select widget
    widget.files_list.value = widget.files_list.options[2]
    assert widget._parent_dir.path == '/'
    assert widget.get_selected_file_path() is None
    assert len(widget.files_list.options) == 1

    with pytest.raises(AttributeError):
        assert widget.get_selected_file_content()


def test_storage_widget(mocker):
    def mockk(token):
        return MockDriveClient()

    mocker.patch('ebrains_drive.connect', mockk)

    if os.environ.get(CLB_AUTH):
        os.environ.pop(CLB_AUTH)

    with pytest.raises(RuntimeError):
        StorageWidget()

    os.environ[CLB_AUTH] = "test_auth_token"
    widget = StorageWidget()

    widget.api.repos_dropdown.value = widget.api.repos_dropdown.options['repo1']
    widget.api.files_list.value = widget.api.files_list.options[1]

    assert widget.get_selected_file_name() == DUMMY_FILENAME
    assert widget.get_selected_file_content() == DUMMY_CONTENT



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import os
from .logger.builder import get_logger

CLB_AUTH = 'CLB_AUTH'

LOGGER = get_logger(__name__)


def get_current_token():
    try:
        from clb_nb_utils import oauth as clb_oauth

        bearer_token = clb_oauth.get_token()
        return bearer_token

    except Exception:
        LOGGER.info("We could not find Collab Auth Token, we will search for env CLB_AUTH variable")

        env_token = os.environ.get(CLB_AUTH)
        if env_token is not None:
            LOGGER.info("We found Collab Auth in environment!")
            return env_token

        raise RuntimeError("Could not authenticate in Collab. Try to define local env  CLB_AUTH or login EBRAINS")



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import json
import setuptools
import shutil
import pkg_resources
from pathlib import Path

with Path('requirements.txt').open() as requirements_txt:
    REQUIRED_PACKAGES = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]

REQUIRED_EXTRA_EXAMPLES = ["tvb-data"]
REQUIRED_EXTRA_TESTS = ["pytest", "pytest-mock"]

HERE = Path(__file__).parent.resolve()

DESCRIPTION = (HERE / "README.md").read_text()

# Get the package info from package.json
pkg_json = json.loads((HERE / "package.json").read_bytes())
version = (
    pkg_json["version"]
        .replace("-alpha.", "a")
        .replace("-beta.", "b")
        .replace("-rc.", "rc")
)

setuptools.setup(name='tvb-widgets',
                 version=version,
                 packages=setuptools.find_packages(),
                 include_package_data=True,
                 install_requires=REQUIRED_PACKAGES,
                 extras_require={"tests": REQUIRED_EXTRA_TESTS, "examples": REQUIRED_EXTRA_EXAMPLES},
                 description=pkg_json['description'],
                 long_description=DESCRIPTION,
                 long_description_content_type="text/markdown",
                 license=pkg_json['license'],
                 author=pkg_json["author"]["name"],
                 author_email=pkg_json["author"]["email"],
                 url=pkg_json["homepage"],
                 keywords='tvb widgets jupyterlab ebrains showcases')

# Cleanup after EGG install. These are created by running setup.py in the source tree
shutil.rmtree('tvb_widgets.egg-info', True)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import inspect
from enum import Enum
from tvb.simulator import integrators


class IntegratorsEnum(Enum):
    BASE_INTEGRATOR = 'Integrator'

    @staticmethod
    def get_integrators_dict():
        integrators_dict = dict()
        for name, obj in inspect.getmembers(integrators):
            if inspect.isclass(obj) and issubclass(obj, integrators.Integrator):
                integrators_dict[obj.__name__] = obj
        return integrators_dict



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

import bz2
import ipywidgets
from tvb.basic.readers import ReaderException

from tvbwidgets.core.exceptions import InvalidFileException
from tvbwidgets.ui.base_widget import TVBWidget
from tvbwidgets.ui.storage_widget import StorageWidget


class TVBWidgetWithBrowser(TVBWidget):
    MSG_TEMPLATE = '<span style="color:{1};">{0}</span>'
    MSG_COLOR = 'red'

    def __init__(self, collab=None, folder=None):
        super().__init__()
        self.storage_widget = StorageWidget(collab, folder)
        self.message_label = ipywidgets.HTML(layout=ipywidgets.Layout(height='25px'))

    def __display_message(self, msg):
        self.message_label.value = self.MSG_TEMPLATE.format(msg, self.MSG_COLOR)

    def __validate_file(self, file_name, accepted_suffix):
        if file_name is None:
            raise InvalidFileException("Please select a file!")

        for suffix in accepted_suffix:
            if file_name.endswith(suffix):
                return suffix
        raise InvalidFileException(
            f"Only {' or '.join(ext for ext in accepted_suffix)} files are supported for this data type!")

    def load_selected_file(self, datatype_cls, accepted_suffix=('.zip',)):
        file_name = self.storage_widget.get_selected_file_name()
        msg = ''
        self.logger.info(f"{file_name}")

        try:
            content_type = self.__validate_file(file_name, accepted_suffix)
        except InvalidFileException as e:
            msg = e.message
            self.logger.error(f"{e}")
            return
        finally:
            self.__display_message(msg)

        content_bytes = self.storage_widget.get_selected_file_content()

        try:
            datatype = datatype_cls.from_bytes_stream(content_bytes, content_type)
            datatype.configure()
            self.add_datatype(datatype)
        except ReaderException as e:
            msg = "The selected file does not contain all necessary data to load this data type! Please check the logs!"
            self.logger.error(f"{e}")
            return
        except Exception as e:
            msg = "Could not load data from this file! Please check the logs!"
            self.logger.error(f"{e}")
            return
        finally:
            self.__display_message(msg)



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#

from .ui.phase_plane_widget import PhasePlaneWidget
from .ui.storage_widget import StorageWidget
from .ui.head_widget import HeadBrowser, HeadWidget, HeadWidgetConfig
from .ui.ts_widget import TimeSeriesWidget, TimeSeriesBrowser
from .ui.pse_widget import PSEWidget
from .ui.pse_launcher_widget import PSELauncher

from IPython.core.display_functions import display
from tvb.datatypes.time_series import TimeSeries


def plot_timeseries(data, sample_freq=None, ch_idx=None, backend='matplotlib'):
    """
    Plots a timeseries widget
    :param data: timeseries data as numpy array or TimeSeries object
    :param sample_freq: float
    :param ch_idx: Channels Index from the max 4D of the data. We assume time is on 0
    :param backend: plotting backend ('matplotlib' or 'plotly'), currently only matplotlib is supported (string)
    """

    if backend == 'matplotlib':
        tsw = TimeSeriesWidget()
        if isinstance(data, TimeSeries):
            tsw.add_datatype(data)
        else:
            tsw.add_data_array(data, sample_freq, ch_idx)
        display(tsw)
    elif backend == 'plotly':
        raise NotImplementedError('Plotly is not currently supported')
    else:
        raise ValueError('The chose backend is not supported, please choose between: "matplotlib" and "plotly"')



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#
import numpy as np
from tvb.basic.neotraits._attr import NArray
from tvbwidgets.core.pse.parameters import launch_local_param
from tvbwidgets.core.pse.parameters import METRICS
from tvbwidgets.ui.base_widget import TVBWidget
from IPython.core.display_functions import display
import ipywidgets as widgets


class PSELauncher(TVBWidget):

    def __init__(self, simulator, connectivity_list=None):
        # type: (Simulator, list) -> None
        super().__init__()
        self.simulator = simulator
        self.connectivity_list = connectivity_list
        self.metrics = METRICS
        self.params_dict = {}
        self.create_dict()
        self.param_1 = None
        self.param_2 = None
        self.warning = None
        self.launch_text_information = None
        self.min_range1 = None
        self.max_range1 = None
        self.step1 = None
        self.min_range2 = None
        self.max_range2 = None
        self.step2 = None
        self.launch_local_button = None
        self.file_name = None
        self.metrics_sm = None
        self.create_informative_texts()
        self.handle_launch_buttons()
        self.create_metrics()
        self.file_options()
        widget = self.create_params()
        display(widget)

    def create_dict(self):
        for elem in type(self.simulator.model).declarative_attrs:
            attribute = getattr(type(self.simulator.model), elem)
            if isinstance(attribute, NArray) and attribute.domain is not None:
                self.params_dict[f"model.{elem}"] = [attribute.domain.lo, attribute.domain.hi, attribute.domain.step]

        for elem in type(self.simulator.coupling).declarative_attrs:
            attribute = getattr(type(self.simulator.coupling), elem)
            if isinstance(attribute, NArray) and attribute.domain is not None:
                self.params_dict[f"coupling.{elem}"] = [attribute.domain.lo, attribute.domain.hi, attribute.domain.step]

        cond_speed_default_value = self.simulator.conduction_speed
        self.params_dict["conduction_speed"] = [0, 10 * cond_speed_default_value, cond_speed_default_value]

        if "noise" in type(self.simulator.integrator).declarative_attrs:
            for elem in type(self.simulator.integrator.noise).declarative_attrs:
                attribute = getattr(type(self.simulator.integrator.noise), elem)
                if isinstance(attribute, NArray) and attribute.domain is not None:
                    self.params_dict[f"integrator.{elem}"] = [attribute.domain.lo, attribute.domain.hi,
                                                              attribute.domain.step]

        if self.connectivity_list is not None:
            self.params_dict["connectivity"] = [0, 0, 0]

    def file_options(self):
        self.file_name = widgets.Text(
            placeholder='Type here',
            description="<b>Name of file</b>",
            disabled=False
        )

    def handle_launch_buttons(self):
        self.launch_local_button = widgets.Button(
            description='Local launch',
            disabled=False,
            button_style='success',
        )

        def local_launch(change):
            self.logger.info("Local launch in progress")
            if self.launch_local_button.button_style == "success":
                self.launch_text_information.value = "<font color='gray'>Local launch in progress"

                if self.param_1.value == "connectivity":
                    x_values = self.connectivity_list
                else:
                    x_values = self._create_input_values(self.min_range1.value, self.max_range1.value, self.step1.value)
                if self.param_2.value == "connectivity":
                    y_values = self.connectivity_list
                else:
                    y_values = self._create_input_values(self.min_range2.value, self.max_range2.value, self.step2.value)
                launch_local_param(self.simulator, self.param_1.value, self.param_2.value, x_values, y_values,
                                   list(self.metrics_sm.value), self.file_name.value)

        self.launch_local_button.on_click(local_launch)

    def create_metrics(self):
        self.metrics_sm = widgets.SelectMultiple(
            options=self.metrics,
            description="<b>Metrics</b>",
            value=[self.metrics[0]],
            disabled=False, layout=widgets.Layout(margin="0px 20px 10px 25px", height="115px", width="340px"))

    def create_informative_texts(self):
        self.warning = widgets.HTML(value="", layout=widgets.Layout(margin="0px 0px 0px 65px"))
        self.launch_text_information = widgets.HTML(value="", layout=widgets.Layout(margin="7px 0px 0px 8px"))

    def _create_input_values(self, min_value, max_value, step):
        return np.arange(min_value, max_value, step).tolist()

    def create_params(self):
        self.param_1 = widgets.Dropdown(
            options=sorted(self.params_dict.keys()),
            description="<b>PSE param1</b>",
            value=list(sorted(self.params_dict.keys()))[0],
            disabled=False
        )
        self.param_2 = widgets.Dropdown(
            options=sorted(self.params_dict.keys()),
            description="<b>PSE param2</b>",
            value=list(sorted(self.params_dict.keys()))[2],
            disabled=False
        )

        def param_changed(change):
            if change['type'] != 'change' or change['name'] != 'value':
                return

            if self.param_1.value != self.param_2.value:
                self.warning.value = ""
                self.launch_local_button.button_style = 'success'
            else:
                self.warning.value = "<b><font color='red'>The parameters should be different!</b>"
                self.launch_local_button.button_style = 'danger'

            if self.param_1.value == change['new'] and self.param_1.value != self.param_2.value:
                if self.param_1.value == "connectivity":
                    self.visibility_range(True, "hidden")
                elif change['old'] == "connectivity":
                    self.visibility_range(True, "visible")

                self.change_range_param(True)
            elif self.param_2.value == change['new'] and self.param_1.value != self.param_2.value:
                if self.param_2.value == "connectivity":
                    self.visibility_range(False, "hidden")
                elif change['old'] == "connectivity":
                    self.visibility_range(False, "visible")

                self.change_range_param(False)
            else:
                if self.param_1.value == "connectivity":
                    self.visibility_range(True, "hidden")
                    self.visibility_range(False, "hidden")
                else:
                    self.visibility_range(True, "visible")
                    self.visibility_range(False, "visible")
                self.change_range_param(True)
                self.change_range_param(False)

        self.param_1.observe(param_changed)
        self.param_2.observe(param_changed)

        return self.pse_params_range()

    def visibility_range(self, param1, visibility):
        if param1:
            self.min_range1.layout.visibility = visibility
            self.max_range1.layout.visibility = visibility
            self.step1.layout.visibility = visibility
        else:
            self.min_range2.layout.visibility = visibility
            self.max_range2.layout.visibility = visibility
            self.step2.layout.visibility = visibility

    def change_range_param(self, param1):
        if param1:
            self.min_range1.value = self.params_dict[self.param_1.value][0]
            self.max_range1.value = self.params_dict[self.param_1.value][1]
            self.step1.value = self.params_dict[self.param_1.value][2]
        else:
            self.min_range2.value = self.params_dict[self.param_2.value][0]
            self.max_range2.value = self.params_dict[self.param_2.value][1]
            self.step2.value = self.params_dict[self.param_2.value][2]

    def pse_params_range(self):
        self.set_range()

        range1 = widgets.VBox(children=[self.min_range1, self.max_range1, self.step1],
                              layout=widgets.Layout(margin="0px 0px 0px 60px"))
        range2 = widgets.VBox(children=[self.min_range2, self.max_range2, self.step2],
                              layout=widgets.Layout(margin="0px 0px 0px 60px"))
        param_box1 = widgets.HBox(children=[self.param_1, range1], layout=widgets.Layout(margin="40px 0px 0px 50px"))
        param_box2 = widgets.HBox(children=[self.param_2, range2], layout=widgets.Layout(margin="30px 0px 0px 50px"))
        buttons_box = widgets.VBox(
            children=[self.launch_local_button, self.launch_text_information],
            layout=widgets.Layout(
                margin="0px 0px 50px 20px"))

        metrics_buttons_box = widgets.HBox(children=[self.metrics_sm, self.file_name, buttons_box],
                                           layout=widgets.Layout(margin="40px "
                                                                        "0px "
                                                                        "30px "
                                                                        "3px"))
        return widgets.VBox(children=[self.warning, param_box1, param_box2, metrics_buttons_box],
                            layout=self.DEFAULT_BORDER)

    def set_range(self):
        self.min_range1 = widgets.FloatText(
            value=self.params_dict[self.param_1.value][0],
            description="<b><font color='gray'>Min range</b>",
            disable=False
        )

        self.max_range1 = widgets.FloatText(
            value=self.params_dict[self.param_1.value][1],
            description="<b><font color='gray'>Max range</b>",
            disable=False
        )

        self.step1 = widgets.FloatText(
            value=self.params_dict[self.param_1.value][2],
            description="<b><font color='gray'>Step</b>",
            disable=False
        )

        self.min_range2 = widgets.FloatText(
            value=self.params_dict[self.param_2.value][0],
            description="<b><font color='gray'>Min range</b>",
            disable=False
        )

        self.max_range2 = widgets.FloatText(
            value=self.params_dict[self.param_2.value][1],
            description="<b><font color='gray'>Max range</b>",
            disable=False
        )

        self.step2 = widgets.FloatText(
            value=self.params_dict[self.param_2.value][2],
            description="<b><font color='gray'>Step</b>",
            disable=False
        )



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2023, TVB Widgets Team
#
import ipywidgets

from tvbwidgets.ui.base_widget import TVBWidget
from tvbwidgets.ui.drive_widget import DriveWidget


class StorageWidget(ipywidgets.Tab, TVBWidget):

    def __init__(self, collab=None, folder=None, **kwargs):
        tab1 = ipywidgets.VBox()
        tab2 = DriveWidget(collab, folder)
        tab3 = ipywidgets.VBox()

        super().__init__([tab1, tab2, tab3], selected_index=1,
                         layout=ipywidgets.Layout(width='550px', height='200px'), **kwargs)

        self.set_title(0, 'Current Selection')
        self.set_title(1, 'Drive')
        self.set_title(2, 'Bucket')
        # TODO uniform API for all tabs
        self.api = tab2

    def get_selected_file_content(self):
        return self.api.get_selected_file_content()

    def get_selected_file_name(self):
        filename = self.api.get_selected_file_path()
        return filename
